# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""``Act@Gory`` encoding and decoding using regular expressions."""

import re

__all__ = ("RegExpActegory",)


class RegExpActegory(object):
    def __init__(self, sep):
        self.sep = sep

    @property
    def sep(self):
        return self._sep

    @sep.setter
    def sep(self, sep):
        self._sep = sep
        self.reset_re()

    # (lb): I considered supporting double-sep as another way to escape it,
    # e.g., `@@` could be same as `\@`. But you run into problems quickly,
    # e.g., 'a@@@b' would be interpreted as 'a\@@b' and never as 'a@\@b',
    # i.e., (act, cat) would be ('a@', 'b') and never ('a', '@b'). Also, the
    # regex is more cumbersome. So we'll just stick with backslash escapes only.

    def reset_re(self):
        self.re_unescaped_sep = re.compile(r"(?<!\\){}".format(self.sep))
        self.esc_sep = r"\\{}".format(self.sep)
        self.raw_sep = r"{}".format(self.sep)

    def escape(self, text):
        escaped = re.sub(self.raw_sep, self.esc_sep, text)
        return escaped

    def unescape(self, text):
        unescaped = re.sub(self.esc_sep, self.raw_sep, text)
        return unescaped

    def split_parts(self, text):
        try:
            escaped_act, escaped_cat = self.re_unescaped_sep.split(text, 1)
        except ValueError:
            return self.unescape(text), None

        activity = self.unescape(escaped_act)
        category = self.unescape(escaped_cat)
        return activity, category



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Fact Editing State Machine"""

from collections import namedtuple
from gettext import gettext as _

from easy_as_pypi_termio.errors import echo_warning
from easy_as_pypi_termio.style import set_coloring, stylize, verify_colors_attrs
from nark.items.activity import Activity
from nark.items.fact import Fact

from ..styling.class_namilize import namilize
from .facts_diff import FactsDiff

__all__ = (
    "FactDressed",
    # PRIVATE:
    #  'FactoidSource',
)


FactoidSource = namedtuple(
    "FactoidSource",
    ("line_num", "line_raw"),
)


class FactDressed(Fact):
    """"""

    def __init__(
        self, *args, dirty_reasons=None, line_num=None, line_raw=None, **kwargs
    ):
        super(FactDressed, self).__init__(*args, **kwargs)
        # For tracking edits between store saves.
        self.dirty_reasons = dirty_reasons or set()
        # For identifying errors in the input.
        self.parsed_source = FactoidSource(line_num, line_raw)
        self.orig_fact = None
        # For Carousel (dob-viewer) navigation.
        self.next_fact = None
        self.prev_fact = None

    # ***

    @property
    def pk_str(self):
        return str(self.pk)

    # ***

    @property
    def short(self):
        friendly = "0x{:12x} / 🏭 {} / {} to {:23} / prev: {:12x} / next: {:12x}".format(
            id(self),
            self.pk is not None and "{:6d}".format(self.pk) or "<None>",
            self.start_fmt_local,
            self.end_fmt_local or "..........now..........",
            self.prev_fact and id(self.prev_fact) or 0,
            self.next_fact and id(self.next_fact) or 0,
        )
        return friendly

    # ***

    def copy(self, *args, **kwargs):
        """ """
        new_fact = super(FactDressed, self).copy(*args, **kwargs)
        new_fact.dirty_reasons = set(list(self.dirty_reasons))
        new_fact.parsed_source = self.parsed_source
        new_fact.orig_fact = self.orig_fact or self
        # SKIP: next_fact, prev_fact.
        return new_fact

    # ***

    def friendly_diff(self, other, formatted=False, **kwargs):
        facts_diff = FactsDiff(self, other, formatted=formatted)
        return facts_diff.friendly_diff(**kwargs)

    # ***

    def friendly_str(self, *args, colorful=False, **kwargs):
        was_coloring = set_coloring(colorful)
        friendly_str = super(FactDressed, self).friendly_str(*args, **kwargs)
        set_coloring(was_coloring)
        return friendly_str

    # ***

    FACTOID_STYLE = {}

    @classmethod
    def register_factoid_style(cls, factoid_style=None):
        """Registers a dictionary of style lists for oid_stylize."""
        for part, styles in factoid_style.items():
            errs = ", ".join(verify_colors_attrs(*styles))
            if errs:
                emsg = _("Unknown colors or attrs for “{}”: {}").format(part, errs)
                echo_warning(emsg)
        # Nonetheless, can still use even if some/all unknown colors/attrs.
        cls.FACTOID_STYLE = factoid_style or {}

    def oid_stylize(self, oid_part, oid_text):
        """Stylizes parts of the Factoid with color and emphasis."""
        try:
            styles = FactDressed.FACTOID_STYLE[oid_part]
        except KeyError:
            styles = []
        stylized = styles and stylize(oid_text, *styles) or oid_text
        return stylized

    # ***

    def squash(self, other, squash_sep=""):
        def _squash():
            # (lb): The squash is a useful end user application feature for existing
            # facts, and I'm not sure what else it might be used for, so I'm putting
            # a bunch of asserts here to force you to re-read this comment when next
            # this code blows up because new usage and you realize you can assuredly
            # delete this comment and one or all of these assert and you will likely
            # be just fine.
            assert other.pk is None or other.pk < 0
            assert not self.deleted
            assert not other.deleted
            assert not other.split_from
            # When squashing, the first fact should have a start, but not an end.
            # And we do not care about other; it could have a start, or an end, or
            # neither.
            assert self.start
            assert not self.end

            self.end = other.start or other.end

            if other.activity_name or other.category_name:
                # (lb): MAYBE: Do we care that this is destructive?
                self.activity = other.activity

            self.tags_replace(self.tags + other.tags)

            description_squash(other, squash_sep)

            self.dirty_reasons.add("squash")
            if self.end:
                self.dirty_reasons.add("stopped")
                self.dirty_reasons.add("end")

            other.deleted = True
            # For completeness, and to make verification easier.
            other.start = self.start
            other.end = self.end

            other.dirty_reasons.add("deleted-squashed")

        def description_squash(other, squash_sep=""):
            if not other.description:
                return
            # (lb): Build local desc. copy, because setter stores None, never ''.
            new_description = self.description or ""
            new_description += squash_sep if new_description else ""
            new_description += other.description
            self.description = new_description
            other.description = None

        _squash()

    # ***

    @classmethod
    def create_from_factoid(cls, factoid, *args, **kwargs):
        """Creates a new Fact from Factoid text, and sets bulk import metadata."""
        new_fact, err = super(FactDressed, cls).create_from_factoid(
            factoid, *args, **kwargs
        )
        if new_fact is not None:
            line_num = 1
            line_raw = factoid
            new_fact.parsed_source = FactoidSource(line_num, line_raw)
        return new_fact, err

    # ***

    @property
    def dirty(self):  # MAYBE: Rename: positive()?
        # MAYBE/FIXME: Set dirty_reasons if fact.pk < 0, on new FactDressed.
        return (self.unstored or len(self.dirty_reasons) > 0) and not self.is_gap

    # *** Linked list methods.

    @property
    def has_next_fact(self):
        return self.next_fact is not None

    @property
    def has_prev_fact(self):
        return self.prev_fact is not None

    # ***

    @property
    def is_gap(self):
        return "interval-gap" in self.dirty_reasons

    @is_gap.setter
    def is_gap(self, is_gap):
        if is_gap:
            self.dirty_reasons.add("interval-gap")
        else:
            self.dirty_reasons.discard("interval-gap")

    @classmethod
    def new_gap_fact(cls, start, end=None, pk=None):
        if pk is None:
            pk = -1
        activity = Activity(name="")
        gap_fact = FactDressed(
            pk=pk,
            activity=activity,
            start=start,
            end=end,
        )
        # Add 'interval-gap' dirty reason.
        gap_fact.is_gap = True
        # Mark deleted until edited, so gap is not saved unless edited.
        gap_fact.deleted = True
        # No exceptions! All Fact copies must eventually lead to the original.
        gap_fact.orig_fact = gap_fact.copy()
        return gap_fact

    # *** Presentation concerns.

    def oid_tags(self, *args, colorful=False, **kwargs):
        was_coloring = set_coloring(colorful)
        friendly_tags = super(FactDressed, self).oid_tags(*args, **kwargs)
        set_coloring(was_coloring)
        return friendly_tags

    # ***

    TAGS_TUPLE_STYLE = {}

    @classmethod
    def register_tags_tuples_style(cls, tags_tuples_style=None):
        """Registers a dictionary of PTK-style styles for tags_tuples."""
        # (lb): See command in CustomHeaderValues. These two special styles,
        # 'value-tag-#' and 'value-tag-label', let you style the hash symbol
        # separately from the tag label. As opposed to the 'value-tag' option.
        cls.TAGS_TUPLE_STYLE = tags_tuples_style or {}

    # NOTE/2020-01-27: (lb): Returns PPT tuple, not something dob knows about.
    # - We could move this method to dob-viewer, to truly decouple.
    #   But then dob-viewer needs to override FactDressed (self.store.fact_cls).
    #   (For now, I'm happy this method at least made it out of nark!)
    def tags_tuples(
        self,
        hashtag_token="#",
        quote_tokens=False,
        colorful=False,
        split_lines=False,
    ):
        def format_tagname(tag):
            tagged = []
            #
            # (lb): I had ' fg: ' prefix for class name, but not needed.
            tclss_fmt = " class:tag-{}".format(namilize(tag.name))
            #
            token_fmt = ""
            token_fmt += fetch_tag_part_fmt("value-tag-#")
            token_fmt += tclss_fmt
            tagged.append((token_fmt, hashtag_token))
            #
            tname_fmt = ""
            tname_fmt += fetch_tag_part_fmt("value-tag-label")
            tname_fmt += tclss_fmt
            tagged.append((tname_fmt, tag.name))
            #
            if quote_tokens:
                fmt_quote = ("", '"')
                tagged.insert(0, fmt_quote)
                tagged.append(fmt_quote)
            return tagged

        def fetch_tag_part_fmt(style_attr):
            if not colorful:
                return ""
            try:
                return FactDressed.TAGS_TUPLE_STYLE[style_attr]
            except KeyError:
                return ""

        # NOTE: The returned string includes leading space if nonempty!
        tagnames = []
        if self.tags:
            # Build array of PPT tuples.
            fmt_sep = ("", "\n") if split_lines else ("", " ")
            n_tag = 0
            for fmtd_tagn in self.tagnames_sorted_formatted(format_tagname):
                if n_tag > 0:
                    tagnames += [fmt_sep]
                n_tag += 1
                tagnames += fmtd_tagn
        return tagnames

    # ***

    def html_notif(self, cut_width_description=None, sep=": "):
        """
        A briefer Fact one-liner using HTML. Useful for, e.g., notifier toast.
        """
        # (lb): To be honest, the only HTML herein is the <i>No activity</i>.
        was_coloring = set_coloring(False)
        duration = "[{}]".format(self.format_delta(style=""))
        actegory = self.oid_actegory(empty_actegory_placeholder="<i>No activity</i>")
        description = self.oid_description(cut_width=cut_width_description, sep=sep)
        simple_str = "{} {}{}".format(
            duration,
            actegory,
            description,
        )
        set_coloring(was_coloring)
        return simple_str

    # ***



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

""""""

# For the fixture to work, either import what you need, or glob it all.
# - Globbing is easier and more akin to how conftest.py normally magically works.
# - Skip the linter's two gripes:
#    noqa: F401 'foo.bar' imported but unused
#    noqa: F403 'from foo.bar import *' used; unable to detect undefined names
from dob_bright.tests.conftest import *  # noqa: F401, F403

pytest_plugins = (
    # Import tmp_appdirs fixture.
    "easy_as_pypi_appdirs.tests.appdirs_mock",
    # Set KeyChainedValue._envvar_prefix.
    "tests.config.envvar_prefix",
    # Call app_dirs.register_application.
    "tests.config.init_app_dirs",
    "nark.tests.backends.sqlalchemy.conftest",
    # Make sure fixtures required by fixtures available, e.g., 'base_config'.
    "nark.tests.conftest",
)



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Controller functionality: Logging features."""

import logging

from nark.helpers import logging as logging_helpers

from .config_manager import Controller_ConfigManager

__all__ = ("Controller_LoggingOutput",)


class Controller_LoggingOutput(
    # For self.config, depends on:
    #  Controller_ConfigManager  # self.config['dev.*' and 'log.*']
    # but for deliberate super() chaining: load data store layer next:
    Controller_ConfigManager,
):
    """Maintains Controller's self.client_logger."""

    def __init__(self, *args, **kwargs):
        super(Controller_LoggingOutput, self).__init__(*args, **kwargs)

    # ***

    def setup_logging(self, verbose=False, verboser=False):
        """Setup logging for the lib_logger as well as client specific logging."""
        self.client_logger = logging.getLogger("dob")
        loggers = self.get_loggers()
        for logger in loggers:
            logger.handlers = []
        # Clear existing Handlers, and set the level.
        # MAYBE: Allow user to specify different levels for different loggers.
        cli_log_level_name = self.config["dev.cli_log_level"]
        cli_log_level, warn_name = logging_helpers.resolve_log_level(cli_log_level_name)
        # We can at least allow some simpler optioning from the command args.
        if verbose:
            cli_log_level = min(logging.INFO, cli_log_level)
        if verboser:
            cli_log_level = min(logging.DEBUG, cli_log_level)
        # 2019-01-25 (lb): I have not had any issues for past few weeks, but,
        #   just FYI in case you do, you might need to clear handlers on
        #   lib_logger and sql_logger, e.g.,:
        #        for logger in loggers:
        #            logger.handlers = []
        #            logger.setLevel(cli_log_level)
        self.client_logger.handlers = []
        self.client_logger.setLevel(cli_log_level)

        color = self.config["log.use_color"]
        formatter = logging_helpers.formatter_basic(color=color)

        if self.config["log.use_console"]:
            console_handler = logging.StreamHandler()
            logging_helpers.setup_handler(console_handler, formatter, *loggers)

        logfile = self.config["log.filepath"]
        if logfile:
            file_handler = logging.FileHandler(logfile, encoding="utf-8")
            logging_helpers.setup_handler(file_handler, formatter, *loggers)

        if warn_name:
            self.client_logger.warning(
                "Unknown Client.cli_log_level specified: {}".format(cli_log_level)
            )

    def get_loggers(self):
        loggers = [
            self.lib_logger,
            self.sql_logger,
            self.client_logger,
        ]
        return loggers

    def bulk_set_log_levels(self, log_level):
        for logger in self.get_loggers():
            logger.setLevel(log_level)

    def disable_logging(self):
        loggers = [
            self.lib_logger,
            self.sql_logger,
            self.client_logger,
        ]
        for logger in loggers:
            logger.handlers = []
            logger.setLevel(logging.NOTSET)

    # ***



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""The Controller."""

import inspect
import os
import sys
from gettext import gettext as _

from easy_as_pypi_termio.echoes import click_echo, highlight_value
from easy_as_pypi_termio.errors import echo_warning, exit_warning
from nark.items.fact import Fact

from .. import help_newbs as help_strings
from .config_manager import Controller_ConfigManager
from .database_store import Controller_DatabaseStore
from .logging_output import Controller_LoggingOutput
from .pdb_subprocess import Controller_PdbSubprocess
from .post_processor import Controller_PostProcessor
from .setup_coloring import Controller_SetupColoring

__all__ = ("Controller",)


# ***
# *** [CONTROLLER] NarkControl Controller.
# ***

# (lb): This was originally one long class... now it's a class conglomeration.
# Both approaches have their pros and cons, the latter being more reusable and
# composable, but at the expense of readability. And it's not like using
# multiple class inheritance decouples this class at all; it just spreads
# to code out to separate files, which could make grokking more difficult.


class Controller(
    # Ordered so earlier items depend on later items, and also
    # to implement cooperative inheritance chaining. Note that
    # we might want to reduce this complexity and make attributes
    # for each of the classes instead, and maybe also shim functions,
    # but right now the Controller acts as a large multifaceted interface.
    Controller_PostProcessor,
    Controller_DatabaseStore,
    Controller_SetupColoring,
    Controller_PdbSubprocess,
    Controller_LoggingOutput,
    Controller_ConfigManager,
):
    """
    Organizes the Controller stack for dob, layering various functionality.

    Nark provides a basic Controller, and then dob adds all the CLI/TUI
    goodies, like config handling, logging, debugging capabilities,
    coloring, a UX interface to the database store, and arbitrary dob
    command post processing (that plugins use to add value to the app).
    """

    def __init__(self, config=None):
        super(Controller, self).__init__(config)

    # ***

    def insist_germinated(self, fact_cls=Fact):
        """Assist user if config or database not present."""

        def _insist_germinated():
            # Check if both self.store_exists and self.looks_like_config.
            if self.is_germinated:
                # We're good!
                self.standup_store(fact_cls)
                return

            if self.has_no_user_files:
                help_newbie_onboard()
            else:
                alert_user_if_config_file_unwell_or_store_absent()
            sys.exit(1)

        def help_newbie_onboard():
            message = help_strings.NEWBIE_HELP_ONBOARDING(self.ctx)
            click_echo(inspect.cleandoc(message), err=True)

        def alert_user_if_config_file_unwell_or_store_absent():
            self.alert_user_if_config_file_unwell()
            if not self.store_exists:
                oblige_user_create_store()

        def oblige_user_create_store():
            message = help_strings.NEWBIE_HELP_CREATE_STORE(
                self.ctx,
                db_path=self.config["db.path"],
                val_source=self.config.asobj.db.path.source,
            )
            click_echo(inspect.cleandoc(message), err=True)

        _insist_germinated()

    # ***

    def create_config_and_store(self, fact_cls=Fact):
        def _create_config_and_store():
            if not self.is_germinated:
                germinate_config_and_store()
            else:
                exit_already_germinated()

        def germinate_config_and_store():
            create_config_maybe()
            create_store_maybe()

        def create_config_maybe():
            cfg_path = self.configurable.config_path
            if not os.path.exists(cfg_path):
                self.create_config(force=False)
            else:
                click_echo(
                    _("Configuration already exists at {}").format(
                        highlight_value(cfg_path),
                    )
                )

        def create_store_maybe():
            # MEH: (lb): If the engine is not SQLite, this function cannot behave
            # like create_config_maybe, which tells the user if the things exists
            # already, because the storage class, SQLAlchemyStore, blindly calls
            # create_all (in create_storage_tables) without checking if db exists.
            skip_standup = self.check_sqlite_store_ready()
            if skip_standup:
                click_echo(self.data_store_exists_at)
            else:
                self._standup_and_version_store(fact_cls)

        def exit_already_germinated():
            exit_warning(
                _("Dob is already setup. Run `{} details` for info.").format(
                    self.arg0name
                )
            )

        _create_config_and_store()

    # ***

    # (lb): MAYBE/2020-12-27: This seems misplaced.

    def find_latest_fact(self, restrict=None):
        try:
            return self.facts.find_latest_fact(restrict=restrict)
        except Exception as err:
            # (lb): Unexpected! This could mean more than one ongoing Fact found!
            echo_warning(str(err))

    def find_oldest_fact(self):
        return self.facts.find_oldest_fact()

    # ***



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import json
import os
from gettext import gettext as _

from easy_as_pypi_appdirs import AppDirs
from nark.config import ConfigRoot
from nark.config.log_levels import (
    get_log_level_safe,
    get_log_name_safe,
    must_verify_log_level,
)
from nark.helpers.parsing import FACT_METADATA_SEPARATORS

__all__ = (
    "DobConfigurableDev",
    "DobConfigurableEditor",
    "DobConfigurableFact",
    "DobConfigurableLog",
    "DobConfigurableTerm",
    # PRIVATE:
    # '_styling_file_path',
)


# ***
# *** Client (dob) Config.
# ***


def _styling_file_path(basename):
    if not AppDirs.is_ready:
        return ""
    dirname = AppDirs().user_config_dir
    subdir = "styling"
    # E.g., /home/user/config/.dob/styling/{basename}
    return os.path.join(dirname, subdir, basename)


@ConfigRoot.section("editor")
class DobConfigurableEditor(object):
    """"""

    def __init__(self, *args, **kwargs):
        # See comment in NarkConfigurable:
        #   Cannot super because decorator shenanigans.
        # DEATH: super(DobConfigurableEditor, self).__init__(*args, **kwargs)
        # FIXME/2019-11-30 17:47: Could test again: (lb): I might have fixed issue.
        pass

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, center the interface in the terminal (on `dob edit`)"),
    )
    def centered(self):
        return False

    # ***

    @property
    @ConfigRoot.setting(
        # FIXME/2019-11-29: (lb): Help Docs: Add command to list dob + pygments.lexers.
        _("dob or Pygments lexer. See: https://pygments.org/docs/lexers/"),
    )
    def lexer(self):
        # (lb): I'm a reSTie, personally, so let's just default to that.
        # (Not to dishonor other markups, I'm friends with markdownies,
        # I even LaTeXeD once in my life, and let's not talk about that
        # one time I was into this really weird markup called Textile.)
        # MAYBE/2019-11-29: (lb): Perhaps default to '' (and move this to *my* config).
        return "RstLexer"

    # ***

    @property
    @ConfigRoot.setting(
        _(
            "UX style: '' or 'default' applies none;"
            "'night', 'light', or 'color' applies builtin;"
            " or name your own (see: styles_fpath)."
        ),
    )
    def styling(self):
        # Do not specify a style (or specify 'default') so that the UX is not
        # styled, but uses the current terminal color scheme.
        return ""

    # ***

    @property
    @ConfigRoot.setting(
        # FIXME/2019-11-30: Backlog entry: how best to split long help string.
        # - One option: Split here with explicit newlines. Works well for:
        #                 dob config dump -T tabulate
        #               but does not work with -T friendly
        #               - We could default to tabulate...
        #                 or maybe help() callback that customizes
        #                 text based on -T option?
        # - Another option: Truncate after first period '.' on global dob-bright-dump,
        #               but include complete help on dob-bright-dump-setting (e.g.,)
        _(
            "Path to file containing newline-separated ‘act@gory’ and ‘#tag’\n"
            "names to omit from completions and suggestions (regex-aware)."
        ),
        name="ignore_fpath",
    )
    def ignore_list_path(self):
        # E.g., /home/user/config/.dob/styling/ignore.list
        return _styling_file_path(basename="ignore.list")

    # ***

    @property
    @ConfigRoot.setting(
        _("Path to file containing UX styles used by “editor.styling”."),
        name="styles_fpath",
    )
    def styles_conf_path(self):
        # E.g., /home/user/config/.dob/styling/styles.conf
        return _styling_file_path(basename="styles.conf")

    # ***

    @property
    @ConfigRoot.setting(
        _("Path to file defining object matching used to stylize UX."),
        name="rules_fpath",
    )
    def rules_conf_path(self):
        # E.g., /home/user/config/.dob/styling/rules.conf
        return _styling_file_path(basename="rules.conf")

    # ***


# ***


@ConfigRoot.section("fact")
class DobConfigurableFact(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    @property
    @ConfigRoot.setting(
        _("Acceptable separator(s) to delimit Fact meta data from description."),
    )
    def separators(self):
        # Rather than `return ''` and act like there are no separators, show
        # the default value that nark uses, so that user are better educated.
        # FYI: json.dumps([",", ":"]) → '[",", ":"]'
        return json.dumps(FACT_METADATA_SEPARATORS)


# ***


@ConfigRoot.section("dev")
class DobConfigurableDev(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    @property
    @ConfigRoot.setting(
        _(
            "The log level for frontend (dob) squaller"
            " (using Python logging library levels)"
        ),
        validate=must_verify_log_level,
        conform=get_log_level_safe,
        recover=get_log_name_safe,
    )
    def cli_log_level(self):
        return "WARNING"


# ***


@ConfigRoot.section("log")
class DobConfigurableLog(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    # MAYBE/2019-11-29: (lb): I'm not sure utility of setting logfile name,
    # but not the path. Perhaps add log.dirname option.
    # - Or
    @property
    @ConfigRoot.setting(
        # MAYBE/2019-11-29: Resolve appdirs path -- or somehow show user full log path.
        # Or, if (hypothetical) log.dirname option added, substitute in help text here.
        _("Filename of dob log under AppDirs.user_log_dir"),
    )
    def filename(self):
        return "dob.log"

    # ***

    # User should not set logfile_path in config, because it's
    # generated, but we make it a pseudo-setting so the code can
    # access just the same as other settings (and the value depends
    # on another config value, the log_filename, so it sorta is a
    # config value, just a derived one).
    @property
    @ConfigRoot.setting(
        _("Generated value."),
        ephemeral=True,
    )
    def filepath(self):
        if self is None:
            # If called on class definition, before kcv._section set, return
            # empty string, which indicates the value type. However, because
            # ephemeral=True, _deduce_value_type won't call this method with
            # self set to None -- meaning, this return statement technically
            # unreachable. But we'll still keep it, for completeness.
            return ""
        log_dir = AppDirs().user_log_dir
        # Note that self is the root ConfigDecorator, not the DobConfigurableLog.
        log_filename = self["filename"]
        return os.path.join(log_dir, log_filename)

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, use color in log files (useful if tailing logs in terminal)"),
    )
    def use_color(self):
        return False

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, send logs to the console, not the file"),
    )
    def use_console(self):
        return True


# ***


@ConfigRoot.section("term")
class DobConfigurableTerm(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    @property
    @ConfigRoot.setting(
        _("The filename suffix to tell EDITOR so it can determine highlighting"),
    )
    def editor_suffix(self):
        return ""

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, show copyright greeting before every command."),
    )
    def show_greeting(self):
        return False

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, use color and font ornamentation in output."),
    )
    def use_color(self):
        return True

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, send application output to terminal pager."),
    )
    def use_pager(self):
        return False

    # ***

    # TESTME/2020-05-17: Does this issue affect all ASCII table formatters?
    #                    What about journal and factoid formats?
    #                    - My original note said 'tabulate' is 'really slow', but I'd
    #                      bet the `for fact in facts` loop in generate_facts_table
    #                      that assembles the table data is probably not very speedy.
    # Dumping lots of records to the display can be really slow, so apply a
    # default upper limit.
    # - Without this option, a simple `dob list facts` would dump the user's
    #   whole dataset, which might not really be expected, or at best would
    #   be annoying for new users who've just imported their 10-years-running
    #   Hamster db, for example.
    # TESTME/2020-05-17: Does `dob export` at least in a reasonable time?
    #                    (I.e., ensure there's a quick way to output all Facts.)
    # NOTE: There's obviously also --limit/--offset, which is more proactive;
    #       this option is more of a safety mechanism.
    @property
    @ConfigRoot.setting(
        _("Limits the number of result rows output to the terminal."),
    )
    def row_limit(self):
        # 2018-05-09: (lb): Anecdotal evidence suggests 2500 is barely tolerable.
        #   Except it overflows my terminal buffer? and hangs it? Can't even
        #   Ctrl-C back to life?? Maybe less than 2500. 1000 seems fine. Why
        #   would user want that many results in their terminal, anyway?
        #   (Maybe if they were redirecting to a file, but dob can check
        #   sys.stdin.isatty and ignore this option if not.)
        return 1001



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import shutil
import sys

from easy_as_pypi_termio.errors import exit_warning
from easy_as_pypi_termio.paging import ClickEchoPager
from easy_as_pypi_termio.style import stylize
from nark.reports.csv_writer import CSVWriter
from nark.reports.ical_writer import ICALWriter
from nark.reports.json_writer import JSONWriter
from nark.reports.tsv_writer import TSVWriter
from nark.reports.xml_writer import XMLWriter

from .factoid_writer import FactoidWriter
from .journal_writer import JournalWriter
from .table_writer import TableWriter
from .tabulate_results import tabulate_results

__all__ = ("render_results",)


def render_results(
    controller,
    results,
    headers=None,
    query_terms=None,
    show_usage=False,
    show_duration=False,
    hide_description=False,
    custom_columns=None,
    output_format="table",
    table_type="texttable",
    max_width=-1,
    row_limit=0,
    output_obj_or_path=None,
    factoid_rule="",
    datetime_format=None,
    duration_fmt=None,
    spark_total=None,
    spark_width=None,
    spark_secs=None,
    show_totals=False,
    hide_totals=False,
    re_sort=False,
):
    """"""

    def _render_results():
        # Send output to the path or object indicated, or stdout or pager.
        output_obj = output_obj_or_path or ClickEchoPager
        writer = fetch_report_writer(output_format, output_obj)
        n_written = prepare_and_render_results(writer)
        return n_written

    # ***

    def fetch_report_writer(output_format, output_obj):
        writer = fetch_report_writer_cls(
            output_format=output_format,
            table_type=table_type,
        )
        must_prepare_output(writer, output_obj)
        return writer

    def must_prepare_output(writer, output_obj):
        try:
            writer.output_setup(
                output_obj=output_obj,
                row_limit=row_limit,
                datetime_format=datetime_format,
                duration_fmt=duration_fmt,
            )
        except Exception as err:
            # I.e., FileNotFoundError, or PermissionError.
            exit_warning(str(err))

    def fetch_report_writer_cls(output_format, table_type):
        writer = None
        if output_format == "csv":
            writer = CSVWriter()
        elif output_format == "tsv":
            writer = TSVWriter()
        elif output_format == "ical":
            writer = ICALWriter()
        elif output_format == "json":
            writer = JSONWriter()
        elif output_format == "xml":
            writer = XMLWriter()
        elif output_format == "factoid":
            colorful = controller.config["term.use_color"]
            factoid_sep = ""
            if factoid_rule:
                # MAYBE/2020-06-16: (lb): Now that max_width truncates, this no
                # longer works as expected. max_width will generally be -1 now...
                # so either caller has to make ruler the length they want, or we
                # need a separate argument here... sorta think leave up to caller.
                rule_mult = 1
                if max_width > 0 and len(factoid_rule) == 1:
                    rule_mult = max_width
                # FIXME: This color should be customizable, eh. #styling
                factoid_sep = stylize(factoid_rule * rule_mult, "indian_red_1c")
            writer = FactoidWriter(
                colorful=colorful,
                cut_width_complete=max_width,
                factoid_sep=factoid_sep,
                show_duration=show_duration,
            )
        elif output_format == "journal":
            # For default `dob report` command, or when otherwise grouping
            # results by day, show a blank line between sections (Days).
            print_blank_line_between_sections = (
                query_terms.group_days
                and query_terms.sort_cols
                and query_terms.sort_cols[0] == "day"
            )
            writer = JournalWriter(
                section_nls=print_blank_line_between_sections,
            )
        elif output_format == "table":
            writer = TableWriter(
                table_type=table_type,
                max_width=restrict_width(max_width),
            )
        else:
            raise Exception("Unknown output_format: {}".format(output_format))
        return writer

    # ***

    def prepare_and_render_results(writer):
        if headers is not None:
            # For list/usage act/cat/tag, already have ready table and headers.
            n_written = writer.write_report(results, headers, tabulation=None)
        elif query_terms.include_stats or writer.requires_table:
            # For reports with stats, post-process results; possibly sort.
            tabulation = prepare_table_and_columns()
            tabn_headers = [repcol.header for repcol in tabulation.repcols]
            n_written = writer.write_report(tabulation.table, tabn_headers, tabulation)
        else:
            # When dumping Facts to a simple format (e.g., CSV), we can write
            # each Fact on the fly and avoid looping through the results (and,
            # e.g., making a namedtuple for each row). (All Facts are still
            # loaded into memory, but it would be unexpected to have a data
            # store larger than 10s of MBs. So our only concern is speed of
            # the operation, not necessarily how much memory it consumes.)
            n_written = writer.write_facts(results)
        return n_written

    # ***

    def prepare_table_and_columns():
        tabulation = tabulate_results(
            controller,
            results,
            row_limit=row_limit,
            query_terms=query_terms,
            show_usage=show_usage,
            show_duration=show_duration,
            show_description=not hide_description,
            custom_columns=custom_columns,
            output_format=output_format,
            datetime_format=datetime_format,
            duration_fmt=duration_fmt,
            spark_total=spark_total,
            spark_width=spark_width,
            spark_secs=spark_secs,
            show_totals=show_totals,
            hide_totals=hide_totals,
            re_sort=re_sort,
        )
        return tabulation

    # ***

    def restrict_width(max_width):
        if max_width is not None and max_width >= 0:
            return max_width
        elif sys.stdout.isatty():
            # MAGIC_NUMBER: Subtract 1 to leave an empty column border on the right.
            return shutil.get_terminal_size()[0] - 1
        else:
            return 80

    # ***

    return _render_results()


# ***



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Factoid writer output format module."""

from .line_writer import LineWriter

__all__ = ("FactoidWriter",)


class FactoidWriter(LineWriter):
    # Note that specifying cut_width_complete makes the output not technically
    # Factoid format, because the complete Fact will not be represented. But
    # it's nonetheless still useful for showing abbreviated output. Just FYI.
    # (lb): Same with show_elapsed, that text will not be parsable, either.
    def __init__(
        self,
        *args,
        colorful=False,
        cut_width_complete=None,
        factoid_sep="",
        show_duration=False,
        **kwargs,
    ):
        super(FactoidWriter, self).__init__(*args, **kwargs)
        self.colorful = colorful
        self.cut_width_complete = cut_width_complete
        self.factoid_sep = factoid_sep
        # TESTME/FIXME/2020-06-18: (lb): I doubt dob-import parses duration correctly.
        self.show_elapsed = show_duration

    def _write_fact(self, idx, fact):
        one_liners = self.cut_width_complete is not None and self.cut_width_complete > 0
        description_sep = "\n\n" if not one_liners else ": "
        # FIXME: (lb): Add UTC support. Currently, all time assumed "local".
        localize = True
        line = fact.friendly_str(
            shellify=False,
            description_sep=description_sep,
            localize=localize,
            include_id=False,
            colorful=self.colorful,
            cut_width_complete=self.cut_width_complete,
            show_elapsed=self.show_elapsed,
        )
        if (not one_liners or self.factoid_sep) and idx > 0:
            self.output_write()
        self.output_write(line)
        if self.factoid_sep:
            self.output_write()
            self.output_write(self.factoid_sep)

    def write_report(self, table, headers, tabulation=None):
        raise NotImplementedError



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import logging
import os
from unittest import mock

import pytest
from easy_as_pypi_appdirs import app_dirs_with_mkdir
from easy_as_pypi_config.urable import ConfigUrable
from nark.config import ConfigRoot, decorate_config
from nark.helpers import logging as logging_helpers

from dob_bright.controllers.app_controller import Controller


class TestSetupLogging(object):
    """Make sure that our logging setup is executed as expected."""

    def test_setup_logging_and_log_level(self, controller):
        """
        Test that library and client logger have log level set according to config.
        """
        controller.setup_logging()
        assert controller.lib_logger.level == (
            logging_helpers.resolve_log_level(controller.config["dev.lib_log_level"])[0]
        )
        assert controller.client_logger.level == (
            logging_helpers.resolve_log_level(controller.config["dev.cli_log_level"])[0]
        )

    def test_setup_logging_log_console_true(self, controller):
        """Ensure if console logging, lib and client have streamhandlers."""
        controller.config["log.use_console"] = True
        controller.setup_logging()
        assert isinstance(
            controller.client_logger.handlers[0],
            logging.StreamHandler,
        )
        assert isinstance(
            controller.client_logger.handlers[1],
            logging.FileHandler,
        )
        assert isinstance(
            controller.lib_logger.handlers[0],
            logging.StreamHandler,
        )
        assert isinstance(
            controller.lib_logger.handlers[1],
            logging.FileHandler,
        )
        assert len(controller.client_logger.handlers) == 2
        assert len(controller.lib_logger.handlers) == 2
        assert controller.client_logger.handlers[0].formatter

    def test_setup_logging_no_logging(self, controller):
        """Make sure that if no logging enabled, our loggers don't have any handlers."""
        controller.setup_logging()
        # Default loggers are set up in ~/.cache/<app>/log/<app>.log
        assert len(controller.lib_logger.handlers) == 1
        assert len(controller.client_logger.handlers) == 1

    def test_setup_logging_log_file_true(self, controller, tmp_appdirs):
        """
        Make sure that if we enable logfile_path, both loggers receive ``FileHandler``.
        """
        controller.config["log.filepath"] = os.path.join(
            tmp_appdirs.user_log_dir,
            "foobar.log",
        )
        controller.setup_logging()
        assert isinstance(
            controller.lib_logger.handlers[0],
            logging.FileHandler,
        )
        assert isinstance(
            controller.client_logger.handlers[0],
            logging.FileHandler,
        )


class TestGetConfig(object):
    """Make sure turning a config instance into proper config dictionaries works."""

    @pytest.mark.parametrize("cli_log_level", ["debug"])
    def test_log_levels_valid(self, cli_log_level, config_instance):
        """
        Make sure *string loglevels* translates to their respective integers properly.
        """
        config_obj = config_instance(cli_log_level=cli_log_level)
        assert config_obj["dev"]["cli_log_level"] == cli_log_level
        config = decorate_config(config_obj)
        assert config["dev"]["cli_log_level"] == 10
        assert config["dev.cli_log_level"] == 10
        assert config.asobj.dev.cli_log_level.value == 10

    @pytest.mark.parametrize("cli_log_level", ["foobar"])
    def test_log_levels_invalid(self, cli_log_level, config_instance, capsys):
        """Test that invalid *string loglevels* raise ``ValueError``."""
        config_obj = config_instance(cli_log_level=cli_log_level)
        with pytest.raises(
            ValueError,
            match=r"^Unrecognized value for setting ‘cli_log_level’: “foobar”.*",
        ):
            _config = decorate_config(config_obj)  # noqa: F841 unused local
        out, err = capsys.readouterr()
        assert out == ""
        assert err == ""

    def test_invalid_store(self, config_instance):
        """Make sure that passing an ORM other than 'sqlalchemy' raises an exception."""
        config_obj = config_instance(orm="foobar")
        match_former = r"Unrecognized value for setting ‘orm’"
        match_latter = r"“foobar” \(Choose from: ‘sqlalchemy’\)"
        with pytest.raises(
            ValueError,
            match=r"^{}: {}$".format(match_former, match_latter),
        ):
            _config = decorate_config(config_obj)  # noqa: F841 unused local

    def test_non_sqlite(self, config_instance):
        """Make sure that passing a postgres config works.

        Albeit actual postgres connections not tested."""
        confnstnc = config_instance(engine="postgres")
        config = decorate_config(confnstnc)
        assert config["db.host"] == confnstnc["db"]["host"]
        assert config["db.port"] == confnstnc["db"]["port"]
        assert config["db.name"] == confnstnc["db"]["name"]
        assert config["db.user"] == confnstnc["db"]["user"]
        assert config["db.password"] == confnstnc["db"]["password"]


# DUNNO/2023-12-21: This looks like a leftover duplicate from
# the 2020-12-15 refactor (when a single dob monolith was split
# into ten separate packages), but it (somehow?) provides 38 lines
# of coverage (though it looks like it only runs EAPP-config code).
# - CXREF: See same-named and similiar class upstream:
#     ~/.kit/py/easy-as-pypi-config/tests/test_urable.py
class TestGetConfigInstance(object):
    def get_configurable(self):
        return ConfigUrable(
            config_root=ConfigRoot,
            # EAPP-config uses:
            #  configfile_envkey=self.EASY_AS_PYPI_CONFIG_CONFIGFILE_ENVKEY,
            configfile_envkey=Controller.DOB_CONFIGFILE_ENVKEY,
        )

    def test_no_file_present(self, tmp_appdirs, mocker):
        # In lieu of testing from completely vanilla account, ensure config file does
        # not exist (which probably exists for your user at ~/.config/dob/dob.conf).
        # NOTE: AppDirs is a module-scope object with immutable attributes, so we
        # need to mock the entire object (i.e., cannot just patch attribute itself).
        app_dirs_mock = mock.Mock()
        app_dirs_mock.configure_mock(user_config_dir="/XXX")
        app_dirs_mock.configure_mock(user_data_dir="/XXX")
        mocker.patch.object(
            app_dirs_with_mkdir, "AppDirsWithMkdir", return_value=app_dirs_mock
        )
        self.configurable = self.get_configurable()
        self.configurable.load_config(configfile_path=None)
        assert len(list(self.configurable.config_root.items())) > 0
        assert self.configurable.cfgfile_exists is False

    def test_file_present(self, config_instance):
        """Make sure we try parsing a found config file."""
        self.configurable = self.get_configurable()
        self.configurable.load_config(configfile_path=None)
        cfg_val = self.configurable.config_root["db"]["orm"]
        assert cfg_val == config_instance()["db"]["orm"]
        assert config_instance() is not self.configurable.config_root

    def test_config_path_getter(self, tmp_appdirs, mocker):
        """Make sure the config target path is constructed to our expectations."""
        # DRY?/2020-01-09: (lb): Perhaps move repeated ConfigUrable code to fixture.
        self.configurable = self.get_configurable()
        mocker.patch.object(self.configurable, "_load_config_obj")
        self.configurable.load_config(configfile_path=None)
        # The config path, e.g., '/home/user/.config/test__dob-bright/app.conf'
        expectation = self.configurable.configfile_path
        self.configurable._load_config_obj.assert_called_with(expectation)



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

""""""

from gettext import gettext as _

from easy_as_pypi_termio.style import attr, bg, fg

__all__ = (
    "conflict_prefix",
    "prepare_log_msg",
)


# ***


def conflict_prefix(prefix):
    return "{}{}{}".format(
        bg("medium_violet_red"),
        prefix,
        attr("reset"),
    )


# ***


def prepare_log_msg(fact_or_dict, msg_content):
    def _prepare_log_msg():
        try:
            line_num = fact_or_dict["line_num"]
            line_raw = fact_or_dict["line_raw"]
        except KeyError:
            line_num = fact_or_dict["parsed_source.line_num"]
            line_raw = fact_or_dict["parsed_source.line_raw"]
        except TypeError:
            line_num = fact_or_dict.parsed_source.line_num
            line_raw = fact_or_dict.parsed_source.line_raw
        line_num = line_num or 0
        line_raw = line_raw or ""
        # NOTE: Using colors overrides logger's coloring, which is great!
        # FIXME: (lb): Replace hardcoding. Assign from styles.conf. #styling
        return _(
            "{}{}{}: {}{}: {}{} / {}{}{}\n\n{}: {}“{}”{}\n\n{}: {}{}{}".format(
                attr("bold"),
                conflict_prefix(_("Problem")),
                attr("reset"),
                fg("dodger_blue_1"),
                _("On line"),
                line_num,
                attr("reset"),
                attr("underlined"),
                msg_content,
                attr("reset"),
                conflict_prefix(_("  Typed")),
                fg("hot_pink_2"),
                line_raw.strip(),
                attr("reset"),
                conflict_prefix(_(" Parsed")),
                fg("grey_78"),
                fact_or_dict,
                attr("reset"),
            )
        )

    return _prepare_log_msg()



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import json
from gettext import gettext as _

from easy_as_pypi_termio.errors import exit_warning
from nark.helpers.parsing import ParserException

from .fact_dressed import FactDressed
from .fix_times import reduce_time_hint

__all__ = ("must_create_fact_from_factoid",)


# ***


def must_create_fact_from_factoid(
    controller,
    factoid,
    time_hint,
    lenient=False,
):
    def _must_create_fact_from_factoid():
        separators = must_prepare_factoid_item_separators(controller)
        use_hint = reduce_time_hint(time_hint)
        try:
            fact, err = FactDressed.create_from_factoid(
                factoid=factoid,
                time_hint=use_hint,
                separators=separators,
                lenient=lenient,
            )
            controller.client_logger.info(str(err)) if err else None
        except ParserException as err:
            msg = _("Oops! {}").format(err)
            controller.client_logger.error(msg)
            exit_warning(msg)
        # This runs for verify_after, not verify_none/"blank time".
        fact_set_start_time_after_hack(fact, time_hint=use_hint)
        return fact

    def must_prepare_factoid_item_separators(controller):
        sep_string = controller.config["fact.separators"]
        if not sep_string:
            return None
        try:
            separators = json.loads(sep_string)
        except json.decoder.JSONDecodeError as err:
            msg = _("The 'separators' config value is not valid JSON: {}").format(err)
            controller.client_logger.error(msg)
            exit_warning(msg)
        return separators

    # FIXME/DRY: See create.py/transcode.py (other places that use "+0").
    #   (lb): 2019-01-22: Or maybe I don't care (not special enough to DRY?).
    #   (lb): 2019-12-06: My knowledge of this code is now so crusty that
    #                     it's too dangerous to consider DRYing anything!
    #                     Especially not without good test coverage first.
    def fact_set_start_time_after_hack(fact, time_hint):
        # FIXME/2019-01-19 13:00: What about verify_next: and verify_then: ???
        #   TESTME: Write more tests first to see if there's really an issue.
        if time_hint != "verify_after":
            return
        assert fact.start is None and fact.end is None
        # (lb): How's this for a hack!?
        fact.start = "+0"

    # ***

    return _must_create_fact_from_factoid()



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Journal writer output format module."""

from click_hotoffthehamster._compat import term_len

from .line_writer import LineWriter

__all__ = ("JournalWriter",)


class JournalWriter(LineWriter):
    def __init__(self, *args, section_nls=False, **kwargs):
        super(JournalWriter, self).__init__(*args, **kwargs)
        self.section_nls = section_nls

    @property
    def requires_table(self):
        return True

    def write_report(self, table, headers, tabulation=None):
        self.curr_section = None
        return super(JournalWriter, self).write_report(table, headers, tabulation)

    def _write_result(self, row, headers, tabulation=None):
        line = ""

        next_section = self.curr_section
        if self.curr_section is None or row[0] != self.curr_section:
            next_section = row[0]
            if self.section_nls and self.curr_section is not None:
                # Emit a blank row-line.
                self.output_write()

            line += row[0]
            line += " " * (tabulation.max_widths[0] - term_len(row[0]))
        else:
            # Omit the first column value when it's the same as the previous row's.
            # Strip Unicode/ASNI control characters to compute whitespace to fill.
            line += " " * tabulation.max_widths[0]

        i_remainder = 1
        line += "  " + "  ".join([str(val) for val in row[i_remainder:]])

        # LATER/2020-06-03: Print formatted text.
        #  from prompt_toolkit import print_formatted_text
        #  print_formatted_text(...)

        self.output_write(line)

        self.curr_section = next_section



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""User configurable interactive editor styling settings loaders."""

import os
from gettext import gettext as _

from easy_as_pypi_config.fileboss import create_configobj
from easy_as_pypi_termio.errors import echo_warning

from . import load_obj_from_internal, style_conf

__all__ = (
    "load_style_classes",
    "load_style_rules",
    "load_rules_conf",
    "resolve_named_style",
    "resolve_path_rules",
    "resolve_path_styles",
    "DEFAULT_STYLE",
)


DEFAULT_STYLE = "default"


def load_style_classes(controller, style_name="", skip_default=False):
    # (lb): It's times like these -- adding a dict to get around scoping
    # when sharing a variable -- that I think a method (load_style_classes)
    # should be a class. But this works for now.
    load_failed = {"styles": False}

    def _load_style_classes():
        named_style = style_name or resolve_named_style(controller.config)
        classes_dict = load_dict_from_styles_conf(named_style)
        style_classes = instantiate_or_try_internal_style(named_style, classes_dict)
        return style_classes

    def load_dict_from_styles_conf(named_style):
        styles_conf, failed = load_styles_conf(controller.config)
        if failed:
            load_failed["styles"] = True
        elif styles_conf and named_style in styles_conf:
            # We could keep as ConfigObj, but not necessary, e.g.:
            #   classes_dict = create_configobj(styles_conf[named_style])
            classes_dict = styles_conf[named_style]
            return classes_dict
        return None

    def instantiate_or_try_internal_style(named_style, classes_dict):
        if classes_dict is not None:
            controller.affirm(isinstance(classes_dict, dict))
            defaults = prepare_base_style(classes_dict)
            update_base_style(named_style, classes_dict, defaults)
            return defaults
        return load_internal_style(named_style)

    def prepare_base_style(classes_dict):
        # Load base-style (e.g., style_conf.default) to ensure
        # all keys present (and defaulted), and then update that.
        base_style = DEFAULT_STYLE
        if "base-style" in classes_dict:
            base_style = classes_dict["base-style"] or "default"
        try:
            # This gets a StylesRoot object created by _create_style_object.
            defaults = getattr(style_conf, base_style)()
        except AttributeError as err:  # noqa: F841
            # Unexpected, because of choices= on base-style @setting def.
            controller.affirm(False)
            defaults = style_conf.default()
        return defaults

    def update_base_style(named_style, classes_dict, defaults):
        try:
            defaults.update_gross(classes_dict)
        except Exception as err:
            msg = _("Failed to load style named “{0}”: {1}").format(
                named_style,
                str(err),
            )
            echo_warning(msg)

    def load_internal_style(named_style):
        # HARDCODED/DEFAULT: style_classes default: 'default' (Ha!).
        # - This style uses no colors, so the UX will default to however
        #   the terminal already looks.
        style_classes_fn = load_obj_from_internal(
            controller,
            obj_name=named_style,
            internal=style_conf,
            default_name=not skip_default and DEFAULT_STYLE or None,
            warn_tell_not_found=not load_failed["styles"],
            config_key=CFG_KEY_ACTIVE_STYLE,
        )
        # The Carousel/`dob edit` path leaves skip_default=False, so it will
        # receive at least the default config. The fetch styles config path/
        # `dob config styles.conf` wants None if nothing found (skip_default=True).
        controller.affirm(skip_default or style_classes_fn is not None)
        return style_classes_fn and style_classes_fn() or None

    # ***

    return _load_style_classes()


# ***


def load_styles_conf(config):
    """Return 2-tuple, the styles.conf ConfigObj, and a bool indicating failure.

    The config object will be None if the path does not exist, or if it failed to
    loaded. Failure will be False if the object was loaded, or if the path does
    not exists; failure is True if the file exists, but ConfigObj failed to load it.
    """

    def _load_styles_conf():
        styles_path = resolve_path_styles(config)
        if not os.path.exists(styles_path):
            return None, False
        return load_dict_from_user_styling(styles_path)

    def load_dict_from_user_styling(styles_path):
        styles_conf = create_configobj(styles_path, errname="styles_conf")
        if styles_conf is None:
            return None, True
        return styles_conf, False

    return _load_styles_conf()


# ***


def load_style_rules(controller):
    def _load_style_rules():
        rules_confobj = try_load_dict_from_user_style_rules()
        return rules_confobj

    def try_load_dict_from_user_style_rules():
        rules_confobj, failed = load_rules_conf(controller.config)
        if rules_confobj is None:
            return None
        compile_eval_rules(rules_confobj)
        return rules_confobj

    def compile_eval_rules(rules_confobj):
        # Each section may optionally contain one code/eval component. Compile
        # it now to check for errors, with the bonus that it's cached for later
        # ((lb): not that you'd likely notice any change in performance with or
        # without the pre-compile).
        for section, rules in rules_confobj.items():
            if "eval" not in rules:
                continue
            try:
                rules["__eval__"] = compile(
                    source=rules["eval"],
                    filename="<string>",
                    # Specifying 'eval' because single expression.
                    # Could use 'exec' for sequence of statements.
                    mode="eval",
                )
            except Exception as err:
                rules_path = resolve_path_rules(controller.config)
                msg = _("compile() failed on 'eval' from “{0}” in “{1}”: {2}").format(
                    section,
                    rules_path,
                    str(err),
                )
                echo_warning(msg)

    # ***

    return _load_style_rules()


# ***


def load_rules_conf(config):
    def _load_rules_conf():
        rules_path = resolve_path_rules(config)
        if not os.path.exists(rules_path):
            return None, False
        return wrap_in_configobj(rules_path)

    def wrap_in_configobj(rules_path):
        rules_confobj = create_configobj(rules_path, errname="rules_conf")
        if rules_confobj is None:
            return None, True
        return rules_confobj, False

    return _load_rules_conf()


# ***

CFG_KEY_ACTIVE_STYLE = "editor.styling"


def resolve_named_style(config):
    return config[CFG_KEY_ACTIVE_STYLE]


# ***

CFG_KEY_STYLES_FPATH = "editor.styles_fpath"


def resolve_path_styles(config):
    return config[CFG_KEY_STYLES_FPATH]


# ***

CFG_KEY_RULESETS_FPATH = "editor.rules_fpath"


def resolve_path_rules(config):
    return config[CFG_KEY_RULESETS_FPATH]



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Controller functionality: Application plugin features."""

from .database_store import Controller_DatabaseStore

__all__ = ("Controller_PostProcessor",)


class Controller_PostProcessor(
    # For self.setup_config, depends on:
    #  Controller_ConfigManager
    # but for deliberate super() chaining: load data store layer next:
    Controller_DatabaseStore,
):
    """Implements the @post_process decorator and calls plugins post-command."""

    POST_PROCESSORS = []

    def __init__(self, *args, **kwargs):
        super(Controller_PostProcessor, self).__init__(*args, **kwargs)

    # ***

    def replay_config(self):
        # Called by ensure_plugged_in after loading plugin config.
        # NOTE: This'll re-print any errors messages from the first setup_config.
        #       Except that color will not have been set the first time, so errors
        #       before will be plain; and errors reprinted here will be colorful.
        self.setup_config(self.configfile_path, *self.config_keyvals)

    # ***

    @staticmethod
    def post_processor(func):
        Controller_PostProcessor.POST_PROCESSORS.append(func)

    @staticmethod
    def _post_process(
        ctx,
        controller,
        fact_facts_or_true,
        show_plugin_error=None,
        carousel_active=False,
    ):
        # facts_or_true is one of:
        # - The saved/edited Fact;
        # - a list of Facts;
        # - or True, on upgrade-legacy.
        for handler in Controller_PostProcessor.POST_PROCESSORS:
            handler(
                ctx,
                controller,
                fact_facts_or_true,
                show_plugin_error=show_plugin_error,
                carousel_active=carousel_active,
            )

    def post_process(
        self,
        controller,
        fact_facts_or_true,
        show_plugin_error=None,
        carousel_active=False,
    ):
        Controller_PostProcessor._post_process(
            self.ctx,
            controller,
            fact_facts_or_true,
            show_plugin_error=show_plugin_error,
            carousel_active=carousel_active,
        )

    # ***



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Controller functionality: Config features."""

import inspect

from easy_as_pypi_config.urable import ConfigUrable
from easy_as_pypi_termio.echoes import click_echo
from nark.control import NarkControl

from .. import help_newbs as help_strings
from ..config import ConfigRoot

__all__ = ("Controller_ConfigManager",)


class Controller_ConfigManager(
    NarkControl,
):
    """This is final (lowermost) class in the Controller() super()-chain."""

    DOB_CONFIGFILE_ENVKEY = "DOB_CONFIGFILE"

    def __init__(self, *args, **kwargs):
        super(Controller_ConfigManager, self).__init__(*args, **kwargs)
        self.configurable = None
        self.configfile_path = None
        self.config_keyvals = None

    # ***

    @property
    def is_germinated(self):
        if not self.looks_like_config:
            return False

        # Nothing upstream, e.g.,:
        #   # return NarkControl.is_germinated.fget(self)
        return True

    @property
    def has_no_user_files(self):
        if not self.configurable.cfgfile_exists:
            return True

        return False

    # ***

    def alert_user_if_config_file_unwell(self):
        def _alert_user_if_config_file_unwell():
            if not self.configurable.cfgfile_exists:
                oblige_user_create_config()
            elif not self.looks_like_config:
                oblige_user_repair_config()

        def oblige_user_create_config():
            cfg_path = self.configurable.config_path
            message = help_strings.NEWBIE_HELP_CREATE_CONFIG(self.ctx, cfg_path)
            click_echo(inspect.cleandoc(message), err=True)

        def oblige_user_repair_config():
            cfg_path = self.configurable.config_path
            message = help_strings.NEWBIE_HELP_REPAIR_CONFIG(self.ctx, cfg_path)
            click_echo(inspect.cleandoc(message), err=True)

        _alert_user_if_config_file_unwell()

    # ***

    def ensure_config(self, ctx, configfile_path, *keyvals):
        if self.configurable is not None:
            return
        # NOTE: This ctx.command.name == 'run', i.e., not the same context
        # passed to the Click command handlers. See: pass_context_wrap.
        self.ctx = ctx
        self.configfile_path = configfile_path
        self.config_keyvals = keyvals
        self.setup_config(configfile_path, *keyvals)
        self.wire_configience()

    def setup_config(self, configfile_path, *keyvals):
        self.configurable = self.setup_config_from_file_and_cli(
            configfile_path,
            *keyvals,
        )

    def setup_config_from_file_and_cli(self, configfile_path, *keyvals):
        configurable = ConfigUrable(
            config_root=ConfigRoot,
            configfile_envkey=Controller_ConfigManager.DOB_CONFIGFILE_ENVKEY,
        )
        configurable.load_config(configfile_path)
        configurable.inject_from_cli(*keyvals)
        return configurable

    # ***

    def create_config(self, force):
        self.configurable.create_config(force=force)
        self.wire_configience()

    @property
    def looks_like_config(self):
        def _looks_like_config():
            if not self.configurable or not self.configurable.cfgfile_exists:
                return False

            return cfgfile_looks_like_config()

        def cfgfile_looks_like_config():
            # What's a reasonable expectation to see if the config file
            # legitimately exists? Check that the file exists? Or parse it
            # and verify one or more settings therein? Let's do the latter,
            # seems more robust. We can check the `store` settings, seems
            # like the most obvious setting to check. In any case, we do
            # this just to tell the user if they need to create a config;
            # the app will run just fine without a config file, because
            # defaults!
            try:
                # SYNC_ME: Referenced by NEWBIE_HELP_REPAIR_CONFIG text.
                self.configurable.config_root.asobj.db.orm.value_from_config
                return True
            except AttributeError:
                return False

        return _looks_like_config()

    def round_out_config(self):
        self.configurable.round_out_config()

    def write_config(self, skip_unset=False):
        self.configurable.write_config(skip_unset=skip_unset)

    # ***

    def wire_configience(self, config_root=None):
        self.config = config_root or self.configurable.config_root
        self.capture_config(self.config)
        self.process_config()

    def process_config(self):
        # Used by derived classes if necessary.
        pass

    # ***



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Creates and populates basic config files or (gracefully) dies trying."""

import os
from gettext import gettext as _

from easy_as_pypi_appdirs.exists_or_mkdirs import must_ensure_file_path_dirred
from easy_as_pypi_termio.echoes import click_echo, highlight_value
from easy_as_pypi_termio.errors import exit_warning

__all__ = ("create_basic_conf",)


def create_basic_conf(object_path, object_name, create_object_file_f, force):
    def _create_basic_conf():
        # Exit if config already exists and not --force.
        exit_if_exists_unless_force(object_path, force)
        # This, too, might exit if, e.g., PermissionError.
        must_ensure_file_path_dirred_or_exit(object_path)
        create_object_file_f(object_path)
        echo_path_created(object_path)

    def exit_if_exists_unless_force(object_path, force):
        path_exists = os.path.exists(object_path)
        if path_exists and not force:
            exit_path_exists(object_path)

    def exit_path_exists(object_path):
        exit_warning(_("{} already at {}").format(object_name, object_path))

    def must_ensure_file_path_dirred_or_exit(object_path):
        try:
            must_ensure_file_path_dirred(object_path)
        except Exception as err:
            msg = _("Cannot create path to {} at {}: {}").format(
                object_name,
                object_path,
                str(err),
            )
            exit_warning(msg)

    def echo_path_created(object_path):
        click_echo(
            _("Initialized basic {} at {}").format(
                object_name,
                highlight_value(object_path),
            )
        )

    _create_basic_conf()



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""ASCII Table writer output format module."""

from easy_as_pypi_termio.ascii_table import generate_table
from nark.reports import ReportWriter

__all__ = ("TableWriter",)


class TableWriter(ReportWriter):
    def __init__(
        self,
        *args,
        table_type="texttable",
        max_width=0,
        **kwargs,
    ):
        super(TableWriter, self).__init__(*args, **kwargs)
        self.table_type = table_type
        self.max_width = max_width

    @property
    def requires_table(self):
        return True

    def write_report(self, table, headers, tabulation=None):
        # SKIP:
        #   super(TableWriter, self).write_report(table, headers, tabulation)
        cols_align = None
        if tabulation is not None:
            cols_align = [repcol.align for repcol in tabulation.repcols]
        generate_table(
            table,
            headers,
            output_obj=self.output_file,
            table_type=self.table_type,
            max_width=self.max_width,
            cols_align=cols_align,
        )
        return len(table)



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from dob_bright.reports.tabulate_results import tabulate_results


class TestGenerateTable(object):
    def test_generate_table(self, controller_with_logging, fact):
        """Make sure the table contains all expected fact data."""
        controller = controller_with_logging
        tabulation = tabulate_results(controller, [fact])
        assert tabulation.table[0].start == fact.start.strftime("%Y-%m-%d %H:%M:%S")
        assert tabulation.table[0].activity == fact.activity.name

    def test_generate_basic_table_column_count(self, controller_with_logging):
        """Make sure the number of table columns matches our expectation."""
        controller = controller_with_logging
        tabulation = tabulate_results(controller, [])
        # MAGIC_NUMBER: A basic query (without grouping or addition stats)
        # creates a table with the 8 following columns:
        #   key, start, end, activity, category, tags, description, duration
        assert len(tabulation.repcols) == 8



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Module to provide u18n friendly help text strings for our controller needs."""

from gettext import gettext as _

from easy_as_pypi_config.fileboss import default_config_path
from easy_as_pypi_termio.echoes import highlight_value
from easy_as_pypi_termio.style import attr, bg, fg

from . import __arg0name__

__all__ = (
    "NEWBIE_HELP_WELCOME",
    "NEWBIE_HELP_ONBOARDING",
    "NEWBIE_HELP_CREATE_CONFIG",
    "NEWBIE_HELP_REPAIR_CONFIG",
    "NEWBIE_HELP_CREATE_STORE",
    # Private:
    #  'common_format',
    #  'section_heading',
)


def common_format():
    # (lb): Eh: This is a method because ENABLE_COLORS is False on startup,
    # and fg and attr will return empty strings if called when module is
    # sourced. So wait until help strings are built and it's known if color
    # or not.
    # FIXME: (lb): Replace hardcoding. Retrieve styles from styles.conf. #styling
    #        - Or probably disable for newbie message: we cannot discern terminal
    #          colors, and user will not have assigned a style, probably.
    common_format = {
        "appname": highlight_value(__arg0name__),
        "rawname": __arg0name__,
        "codehi": fg("turquoise_2"),
        "reset": attr("reset"),
        "bold": attr("bold"),
        "italic": attr("italic"),
        "underlined": attr("underlined"),
        "wordhi": fg("chartreuse_3a"),
        "errcol": bg("red_1") + attr("bold"),
    }
    return common_format


# ***
# *** [INDUCTEE] help.
# ***


def NEWBIE_HELP_WELCOME(ctx):
    # FIXME: (lb): Need to test colors on light vs. night terminals. #styling
    _help = (
        _(
            """
        {color}┏━━━━━━━━━━━━━━━━━┓{reset}
        {color}┃ Welcome to dob! ┃{reset}
        {color}┗━━━━━━━━━━━━━━━━━┛{reset}
        """
        )
        .rstrip()
        .strip("\n")
        .format(
            # FIXME: (lb): Replace hardcoding. Assign from styles.conf. #styling
            color=(fg("spring_green_2a") + attr("bold")),
            **common_format(),
        )
    )
    return _help


def section_heading(title):
    return (
        (
            """
        {color}{title}{reset}
        {line_color}{sep:{sep}<{len_title}}{reset}
        """
        )
        .strip()
        .format(
            title=title,
            sep="─",
            len_title=len(title),
            line_color="",
            color="",
            **common_format(),
        )
    )


def NEWBIE_HELP_ONBOARDING(ctx):
    # NOTE: This help is not automatically formatted like other text.
    _help = (
        _(
            """
        {banner}

        Let’s get you setup!

        {init_title}
        To create a fresh, empty database, run:

          {cmd_color}{rawname} init{reset}

        {upgrade_title}
        To learn how to import from the old hamster app,
        or to import an existing dob database, run:

          {cmd_color}{rawname} migrate -h{reset}

        {demo_title}
        If you’d like to get your hands dirty, you can demo
        the application with some example data that you can
        follow as a walk-through. Run:

          {cmd_color}{rawname} demo{reset}
        \b
        """
        )
        .lstrip()
        .format(
            banner=NEWBIE_HELP_WELCOME(ctx),
            upgrade_title=section_heading(_("Import existing facts")),
            init_title=section_heading(_("Start from scratch")),
            demo_title=section_heading(_("Demo Dob")),
            help_title=section_heading(_("Such Help")),
            # FIXME: (lb): Replace hardcoding. Assign from styles.conf. #styling
            cmd_color=fg("spring_green_2a"),
            **common_format(),
        )
    )
    return _help


def NEWBIE_HELP_CREATE_CONFIG(ctx, cfg_path):
    controller = ctx.obj

    _help = (
        _(
            """
        {errcol}ERROR: No config file found at: “{cfg_path}”{reset}

        Where's your config file??

        Verify and correct the configuration file path.

        The configuration file defaults to:

            {default_config_path}

        but you can override it using an environ:

            {envkey}=PATH

        or by specifying a global option:

            -F/--configfile PATH

        If you are certain the path is correct and you want to create
        a new configuration file at the path specified, run init, e.g.,:

            {rawname} -F "{cfg_path}" init
        """
        )
        .strip()
        .format(
            # FIXME/2019-11-19 14:42: Make wrapper for format()
            #                         with common colors defined.
            # - Maybe change errors to white on red, like here,
            #   but only for white on black terms (based on some setting?).
            cfg_path=cfg_path,
            default_config_path=default_config_path(),
            envkey=controller.DOB_CONFIGFILE_ENVKEY,
            **common_format(),
        )
    )
    return _help


def NEWBIE_HELP_REPAIR_CONFIG(ctx, cfg_path):
    controller = ctx.obj

    _help = (
        _(
            """
        {errcol}ERROR: Please fix the config file found at: “{cfg_path}”{reset}

        The configuration file is at least missing the db.orm setting, if not others.

        Either edit and repair the confirguration file manually, or blow it away:
          \b
          {codehi}{rawname} config create --force{reset}
        """
        )
        .strip()
        .format(
            cfg_path=cfg_path,
            default_config_path=default_config_path(),
            envkey=controller.DOB_CONFIGFILE_ENVKEY,
            **common_format(),
        )
    )
    return _help


def NEWBIE_HELP_CREATE_STORE(ctx, db_path, val_source):
    _help = (
        _(
            """
        {errcol}ERROR: No database file found.{reset}

        - There was no file found at: {db_path}

        - The db.path value was set from the ‘{val_source}’ source.

        - For help on configuring settings, try the config help:
          \b
          {codehi}{rawname} config --help{reset}
        """
        )
        .strip()
        .format(db_path=db_path, val_source=val_source, **common_format())
    )
    return _help



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Applies Styles to components that do not have config access."""

import re

from ..crud.fact_dressed import FactDressed
from ..crud.facts_diff import FactsDiff
from .load_styling import load_style_classes

__all__ = ("pre_apply_style_conf",)


def pre_apply_style_conf(controller):
    def _pre_apply_style_conf():
        style_conf = load_style_classes(controller)
        register_factoid_style(style_conf)
        register_facts_diff_style(style_conf)
        register_tags_tuples_style(style_conf)
        # Cache the style conf for the Carousel (if `dob edit` being called). #PROFILING
        controller.style_conf = style_conf
        return style_conf

    # ***

    RE_COMMA_WHITESPACE = re.compile(r",\s*")

    def unpack(csv):
        return RE_COMMA_WHITESPACE.split(csv)

    # ***

    def register_factoid_style(style_conf):
        factoioid_style = unpack_factoioid_style(style_conf)
        FactDressed.register_factoid_style(factoioid_style)

    def unpack_factoioid_style(style_conf):
        factoioid_style = {
            "pk": unpack(style_conf["factoid-pk"]),
            "act@gory": unpack(style_conf["factoid-act@gory"]),
            "#": unpack(style_conf["factoid-#"]),
            "tag": unpack(style_conf["factoid-tag"]),
            "#tag": unpack(style_conf["factoid-#tag"]),
            "start": unpack(style_conf["factoid-start"]),
            "end": unpack(style_conf["factoid-end"]),
            "at": unpack(style_conf["factoid-at"]),
            "to": unpack(style_conf["factoid-to"]),
            "duration": unpack(style_conf["factoid-duration"]),
        }
        return factoioid_style

    # ***

    def register_facts_diff_style(style_conf):
        facts_diff_style = unpack_facts_diff_style(style_conf)
        FactsDiff.register_facts_diff_style(facts_diff_style)

    def unpack_facts_diff_style(style_conf):
        facts_diff_style = {
            "value-diff-old-raw": unpack(style_conf["value-diff-old-raw"]),
            "value-diff-old-ptk": style_conf["value-diff-old-ptk"],
            "value-diff-new-raw": unpack(style_conf["value-diff-new-raw"]),
            "value-diff-new-ptk": style_conf["value-diff-new-ptk"],
        }
        return facts_diff_style

    # ***

    def register_tags_tuples_style(style_conf):
        tags_tuples_style = unpack_tags_tuples_style(style_conf)
        FactDressed.register_tags_tuples_style(tags_tuples_style)

    def unpack_tags_tuples_style(style_conf):
        tags_tuples_style = {
            "value-tag-#": style_conf["value-tag-#"],
            "value-tag-label": style_conf["value-tag-label"],
        }
        return tags_tuples_style

    return _pre_apply_style_conf()



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import shutil
from gettext import gettext as _

from easy_as_pypi_termio.errors import echo_warning
from easy_as_pypi_termio.style import stylize, verify_colors_attrs
from nark.helpers.format_text import format_value_truncate
from nark.helpers.objects import resolve_attr_or_method

__all__ = ("FactsDiff",)


class FactsDiff(object):
    """"""

    def __init__(self, orig_fact, edit_fact, formatted=False):
        self.orig_fact = orig_fact
        self.edit_fact = edit_fact
        self.formatted = formatted
        self.include_newlines = False
        self.exclude_attrs = None

    def __str__(self):
        return "FactsDiff:\n- orig: {}\n- edit: {}".format(
            self.orig_fact.short,
            self.edit_fact.short,
        )

    # ***

    def friendly_diff(
        self,
        truncate=False,
        exclude=None,
        show_elapsed=False,
        show_midpoint=False,
        show_now=False,
    ):
        def _friendly_diff():
            self.include_newlines = True
            self.exclude_attrs = exclude

            result = "" if not self.formatted else []
            result = assemble_diff_attrs(result)

            self.include_newlines = False
            self.exclude_attrs = None

            if not self.formatted:
                result = result.rstrip()
            else:
                while (len(result) > 0) and (not result[-1][1].strip()):
                    result.pop()

            return result

        def assemble_diff_attrs(result):
            result += self.diff_line_assemble(
                None,
                self.time_humanize(show_now),
                "interval",
            )
            if show_midpoint:
                result += self.diff_line_assemble(
                    None,
                    self.time_midpoint(),
                    "midpoint",
                )
            if show_elapsed:
                self_val, other_val = self.diff_time_elapsed(show_now)
                result += self.diff_line_assemble(
                    self_val,
                    other_val,
                    "duration",
                )
            result += self.diff_attrs("start_fmt_local", "start")
            if not show_now:
                result += self.diff_attrs("end_fmt_local", "end")
            else:
                result += self.diff_attrs("end_fmt_local_nowwed", "end")
            if (not truncate) or self.orig_fact.pk or self.edit_fact.pk:
                result += self.diff_attrs("pk_str", "id", beautify=self.beautify_pk)
            result += self.diff_attrs("deleted", "deleted")
            # MAYBE?: (lb): Would we even want to show the split_from fact?
            #  result += self.diff_attrs('split_from', 'split_from')
            result += self.diff_attrs("activity_name", "activity")
            result += self.diff_attrs("category_name", "category")
            if not self.formatted:
                # For the command line, ANSI escape sequences.
                result += self.diff_attrs("oid_tags", "tags", colorful=True)
            else:
                # For the interactive editor/Carousel, PTK style tuples.
                result += self.diff_attrs("tags_tuples", "tags", colorful=True)
            result += self.diff_attrs("description", "description", truncate=truncate)
            return result

        # ***

        return _friendly_diff()

    # ***

    def diff_attrs(
        self,
        prop,
        name=None,
        style_class="",
        truncate=False,
        beautify=None,
        mouse_handler=None,
        **kwargs
    ):
        if (self.exclude_attrs is not None) and (name in self.exclude_attrs):
            return ""
        self_val = resolve_attr_or_method(self.orig_fact, prop, **kwargs)
        other_val = ""
        if self.edit_fact is not None:
            other_val = resolve_attr_or_method(self.edit_fact, prop, **kwargs)
            if callable(other_val):
                other_val = other_val()
            self_val, other_val = self.diff_values_enhance(
                self_val,
                other_val,
                style_class=style_class,
                truncate=truncate,
                beautify=beautify,
                mouse_handler=mouse_handler,
            )
        elif truncate:
            self_val = self.format_value_truncate(self_val)
            self_val = self.format_prepare(
                self_val,
                style_class=style_class,
                mouse_handler=mouse_handler,
            )
            other_val = self.format_prepare(other_val, style_class=style_class)
        attr_diff = self.diff_line_assemble(self_val, other_val, name)
        return attr_diff

    def diff_line_assemble(self, self_val, other_val, name=None):
        prefix = self.diff_values_padded_prefix(name)
        if not self.formatted:
            return self.diff_line_inline_style(self_val, other_val, prefix)
        else:
            return self.diff_line_tuples_style(self_val, other_val, prefix)

    def diff_values_enhance(
        self,
        self_val,
        other_val,
        style_class="",
        truncate=False,
        beautify=None,
        mouse_handler=None,
    ):
        differ = False
        if self_val != other_val:
            differ = True
        if truncate:
            self_val = self.format_value_truncate(self_val)
            other_val = self.format_value_truncate(other_val)
        if beautify is not None:
            self_val, other_val = beautify(self_val, other_val)
            if self_val != other_val:
                differ = True
        if differ:
            self_val = self.format_edited_before(self_val, style_class)
            self_val, other_val = self.format_edited_after(
                self_val,
                other_val,
                style_class,
            )
        else:
            self_val = self.format_prepare(
                self_val,
                style_class=style_class,
                mouse_handler=mouse_handler,
            )
            other_val = self.format_prepare("", style_class=style_class)
        return (self_val, other_val)

    def format_prepare(self, some_val, style_class="", mouse_handler=None):
        if not self.formatted or not isinstance(some_val, str):
            # tags, e.g.,:
            #   [('fg: #C6C6C6 underline', '#'), ('fg: #D7FF87 underline', 'my-tag')]
            if (
                (mouse_handler is not None)
                and (isinstance(some_val, list))
                and (len(some_val) > 0)
                and (isinstance(some_val[0], tuple))
                and (len(some_val[0]) == 2)
            ):
                return [
                    (
                        style_class + tup[0],
                        tup[1],
                        mouse_handler,
                    )
                    for tup in some_val
                ]
            return some_val
        if mouse_handler is None:
            return [(style_class, some_val)]
        return [(style_class, some_val, mouse_handler)]

    def format_value_truncate(self, val):
        # MAGIC_NUMBER: (lb): A third of the terminal (1 / 3.).
        # MAYBE/2019-02-15: Should have Carousel tells us width.
        term_width = shutil.get_terminal_size()[0]
        trunc_width = int(term_width * (1 / 3.0))
        return format_value_truncate(val, trunc_width)

    # ***

    def diff_time_elapsed(self, show_now=False, style_class=""):
        self_val = self.time_elapsed(self.orig_fact, show_now)
        other_val = self.time_elapsed(self.edit_fact, show_now)
        if not self_val:
            # Make 'em the same, i.e., show no diff, no styling.
            self_val = other_val
        return self.diff_values_enhance(self_val, other_val, style_class=style_class)

    def time_elapsed(self, fact, show_now=False):
        # NOTE: start and/or end might be string; e.g., clock or rel. time.
        if (not fact.times_ok) and (not show_now):
            return None
        time_val = fact.format_delta(style="HHhMMm")
        return time_val

    def time_midpoint(self):
        return self.format_prepare(self.edit_fact.time_of_day_midpoint)

    def time_humanize(self, show_now=False):
        return self.format_prepare(
            self.edit_fact.time_of_day_humanize(show_now=show_now)
        )

    def beautify_pk(self, self_val, other_val):
        # (lb): NOTE: This is the only dirty_reasons usage in nark
        #               (most of its usage is in dob).
        # 'lsplit' and 'rsplit' are used by fix_times.resolve_overlapping.
        if "lsplit" in self.edit_fact.dirty_reasons:
            other_val = "New split fact, created before new fact"
        if "rsplit" in self.edit_fact.dirty_reasons:
            other_val = "New split fact, created after new fact"
        return (self_val, other_val)

    # ***

    FACTS_DIFF_STYLE = {}

    @classmethod
    def register_facts_diff_style(cls, facts_diff_style=None):
        """Registers a dictionary of style lists for FactsDiff methods."""
        for part, styles in facts_diff_style.items():
            if not part.endswith("-raw"):
                continue
            errs = ", ".join(verify_colors_attrs(*styles))
            if errs:
                emsg = _("Unknown colors or attrs for “{}”: {}").format(part, errs)
                echo_warning(emsg)
        # Nonetheless, can still use even if some/all unknown colors/attrs.
        cls.FACTS_DIFF_STYLE = facts_diff_style or {}

    @classmethod
    def fetch_style(cls, style_attr):
        try:
            return cls.FACTS_DIFF_STYLE[style_attr]
        except KeyError:
            if style_attr.endswith("-raw"):
                # Expect list of (color, *attrs) to pass to stylize()
                return []
            else:
                # Except ready-to-go PTK tuple style string.
                return ""

    # ***

    def format_edited_before(self, before_val, style_class):
        if not self.formatted:
            styles = FactsDiff.fetch_style("value-diff-old-raw")
            return styles and stylize(before_val, *styles) or before_val
        style = style_class
        style += FactsDiff.fetch_style("value-diff-old-ptk")
        before_parts = []
        if isinstance(before_val, str):
            before_parts += [(style, before_val)]
        elif before_val is not None:
            for tup in before_val:
                before_parts.append((style, tup[1]))
        return before_parts

    def format_edited_after(self, self_val, other_val, style_class):
        if not self.formatted:
            styles = FactsDiff.fetch_style("value-diff-new-raw")
            return (
                "{} | was: ".format(
                    styles and stylize(other_val, *styles) or other_val
                ),
                self_val,
            )
        style = style_class
        style += FactsDiff.fetch_style("value-diff-new-ptk")
        after_parts = []
        if isinstance(other_val, str):
            after_parts += [(style, other_val)]
        elif other_val is not None:
            for tup in other_val:
                after_parts.append((style, tup[1]))
        # (lb): Swap the order, for display purposes.
        #   (These formatting functions are so janky!)
        if self_val and self_val[0][1]:
            after_parts += [("italic", " | was: ")]
        return after_parts, self_val

    # ***

    def diff_values_padded_prefix(self, name):
        if name is None:
            return ""
        prefix_prefix = "  "
        padded_prefix = "{}{:.<19} : ".format(prefix_prefix, name)
        return padded_prefix

    def diff_line_inline_style(self, self_val, other_val, prefix=""):
        format_inline = "{}{}{}".format(prefix, self_val or "", other_val or "")
        format_inline += "\n" if self.include_newlines else ""
        return format_inline

    def diff_line_tuples_style(self, self_val, other_val, prefix="", style_class=""):
        format_tuples = []
        if prefix:
            format_tuples += [(style_class, prefix)]
        if self_val:
            format_tuples += self_val
        if other_val:
            format_tuples += other_val
        if self.include_newlines:
            format_tuples += [("", "\n")]
        return format_tuples



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Prints styles config to stdout to help user setup custom styling."""

from gettext import gettext as _

from config_decorator.config_decorator import ConfigDecorator
from easy_as_pypi_config.dec_wrap import decorate_and_wrap
from easy_as_pypi_config.fileboss import create_configobj, echo_config_obj
from easy_as_pypi_termio.errors import echo_warning

from .load_styling import load_style_classes, load_styles_conf, resolve_path_styles
from .style_conf import KNOWN_STYLES

__all__ = ("echo_styles_conf",)


def echo_styles_conf(controller, style_name="", internal=False, complete=False):
    """Prints style config section(s) from styles.conf or internal sources."""
    config = controller.config

    def _echo_styles_conf():
        if not internal:
            config_obj = load_config_obj()
        else:
            config_obj = load_style_conf()
        if config_obj:
            echo_config_obj(config_obj)
        # Else, already printed error message.

    # ***

    def load_config_obj():
        config_obj, failed = load_styles_conf(config)
        if config_obj:
            return filter_config_obj(config_obj)
        if failed:
            # load_styles_conf prints a ConfigObj error message. Our job is done.
            return None
        return echo_error_no_styles_conf()

    def filter_config_obj(config_obj):
        if not style_name:
            return config_obj
        new_config = create_configobj(conf_path=None)
        try:
            new_config.merge({style_name: config_obj[style_name]})
        except KeyError:
            return echo_error_no_styles_section()
        else:
            return new_config

    def echo_error_no_styles_conf():
        msg = _("No styles file at: {0}").format(resolve_path_styles(config))
        echo_warning(msg)
        return None

    def echo_error_no_styles_section():
        msg = _("No matching section “{0}” found in styles file at: {1}").format(
            style_name,
            resolve_path_styles(config),
        )
        echo_warning(msg)
        return None

    # ***

    def load_style_conf():
        if style_name:
            return load_single_style()
        return load_known_styles()

    def load_single_style():
        style_classes = load_style_classes(controller, style_name, skip_default=True)
        if not style_classes:
            # load_obj_from_internal will have output a warning message.
            return None
        return decorate_and_wrap(style_name, style_classes, complete=complete)

    def load_known_styles():
        """Adds all internal styles to a configobj.

        Includes all settings for the first style ('default'), but only
        those settings that are explicitly set for the remaining styles.
        """
        config_obj = create_configobj(conf_path=None)
        is_default = True
        for name in KNOWN_STYLES:
            style_classes = load_style_classes(controller, name, skip_default=True)
            styles_conf = ConfigDecorator(object, cls_or_name="", parent=None)
            styles_conf.set_section(name, style_classes)
            config_obj.merge(
                styles_conf.as_dict(
                    skip_unset=not is_default and not complete,
                    keep_empties=not is_default and not complete,
                )
            )
            is_default = False
        return config_obj

    # ***

    return _echo_styles_conf()



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from config_decorator.config_decorator import ConfigDecorator
from easy_as_pypi_config.dec_wrap import decorate_and_wrap
from easy_as_pypi_termio.echoes import click_echo, highlight_value
from easy_as_pypi_termio.style import attr

from ..config.config_table import echo_config_decorator_table
from ..crud.interrogate import run_editor_safe
from .create_conf import create_basic_conf
from .load_styling import (
    DEFAULT_STYLE,
    load_style_classes,
    load_styles_conf,
    resolve_named_style,
    resolve_path_styles,
)
from .style_conf import KNOWN_STYLES
from .styles_dump import echo_styles_conf as dump_styles_conf

__all__ = (
    "create_styles_conf",
    "echo_styles_conf",
    "echo_styles_list",
    "echo_styles_table",
    "edit_styles_conf",
)


# *** [CONF] STYLES


def echo_styles_conf(controller, name, internal=False, complete=False):
    dump_styles_conf(controller, name, internal, complete)


# *** [CREATE] STYLES


def create_styles_conf(controller, name, force):
    def _create_styles_conf():
        # SIMILAR funcs: See also: ConfigUrable.create_config and
        #   reset_config; and styles_cmds.create_styles_conf;
        #                  and rules_cmds.create_rules_conf.
        object_name = _("Styles file")
        styles_path = resolve_path_styles(controller.config)
        create_basic_conf(styles_path, object_name, create_styles_file, force)

    def create_styles_file(styles_path):
        # Load specified style, or DEFAULT_STYLE if not specified.
        style_name = name or DEFAULT_STYLE
        style_classes = load_style_classes(controller, style_name=style_name)
        config_obj = decorate_and_wrap(style_name, style_classes, complete=True)
        config_obj.filename = styles_path
        config_obj.write()

    _create_styles_conf()


# *** [EDIT] STYLES


def edit_styles_conf(controller):
    styles_path = resolve_path_styles(controller.config)
    run_editor_safe(filename=styles_path)


# *** [LIST] STYLES


def echo_styles_list(controller, internal=False):
    """"""

    def _echo_styles_list():
        active_style = resolve_named_style(controller.config)
        print_styles_list(KNOWN_STYLES, active_style, _("Built-in styles"))
        if not internal:
            user_styles = fetch_user_styles_list()
            print_styles_list(user_styles, active_style, _("After-market styles"))

    def fetch_user_styles_list():
        config_obj, failed = load_styles_conf(controller.config)
        if config_obj:
            return config_obj.keys()
        return []

    def print_styles_list(styles_list, active_style, title):
        click_echo("{}{}{}".format(attr("underlined"), title, attr("reset")))
        for style_name in styles_list:
            if style_name == active_style:
                sparkler = "* "
                ornament = attr("bold")
            else:
                sparkler = "  "
                ornament = ""
            click_echo("{}{}".format(ornament, highlight_value(sparkler + style_name)))

    return _echo_styles_list()


# *** [SHOW] STYLES


def echo_styles_table(controller, name, output_format):
    def _echo_styles_table():
        style_name = name or resolve_named_style(controller.config)
        style_classes = load_style_classes(
            controller,
            style_name=style_name,
            skip_default=True,
        )
        if style_classes is not None:
            echo_table(style_name, style_classes)
        # Else, already printed error.

    def echo_table(style_name, style_classes):
        condec = ConfigDecorator.create_root_for_section(style_name, style_classes)
        conf_objs = [condec]
        echo_config_decorator_table(
            cfg_decors=conf_objs,
            exclude_section=False,
            # Passed on to render_results:
            controller=controller,
            output_format=output_format,
        )

    _echo_styles_table()



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Controller functionality: pdb-related developer-friendly features."""

from gettext import gettext as _

from .. import __arg0name__
from .logging_output import Controller_LoggingOutput

__all__ = ("Controller_PdbSubprocess",)


class Controller_PdbSubprocess(
    # For self.config, depends on:
    #  Controller_ConfigManager  # self.config['dev.catch_errors']
    # For self.client_logger, depends on.
    #  Controller_LoggingOutput,
    # but for deliberate super() chaining: load data store layer next:
    Controller_LoggingOutput,
):
    """Exposes Controller.affirm and Controller.pdb_set_trace.

    - Especially useful for debugging TUI applications.
    """

    def __init__(self, *args, **kwargs):
        super(Controller_PdbSubprocess, self).__init__(*args, **kwargs)

        # Not necessarily pdb-related, but doesn't fit in any existing Controller
        # sub-class (so this class is as good as any a container for it).
        self.arg0name = __arg0name__

    # ***

    def affirm(self, condition, message=None):
        if condition:
            return
        self.client_logger.error(_("Dob detected an Illegal State!"))

        append_context = ""
        if message:
            append_context = f": {message}"

        self.client_logger.error(_("Dob detected an Illegal State") + append_context)
        if not self.config["dev.catch_errors"]:
            return
        import traceback

        traceback.print_stack()
        traceback.print_exc()
        self.pdb_set_trace()

    # ***

    def pdb_set_trace(self):
        import pdb

        self.pdb_break_enter()
        pdb.set_trace()
        self.pdb_break_leave()

    def pdb_break_enter(self):
        import subprocess

        # If the developer breaks into code from within the Carousel,
        # i.e., from within the Python Prompt Toolkit library, then
        # pdb terminal echo of stdin back to stdout is broken. You'll
        # see that pdb.stdin and pdb.stdout still match the sys.__stdin__
        # and sys.__stdout__, so that's not the issue -- it's that pdb
        # terminal is in *raw* mode. We can fix this by shelling to stty.
        proc = subprocess.Popen(["stty", "--save"], stdout=subprocess.PIPE)
        (stdout_data, stderr_data) = proc.communicate()
        self.stty_saved = stdout_data.strip()
        # Before breaking, twiddle the terminal away from PPT temporarily.
        subprocess.Popen(["stty", "sane"])

    def pdb_break_leave(self):
        import subprocess

        # Aha! This is awesome! We can totally recover from an interactive
        # debug session! First, restore the terminal settings (which we
        # reset so the our keystrokes on stdin were echoed back to us)
        # so that sending keys to PPT works again.
        subprocess.Popen(["stty", self.stty_saved])
        # And then the caller, if self.carousel, will redraw the interface
        # (because it has the handle to the application).
        self.client_logger.debug(_("Get on with it!"))

    # ***



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import os
import platform
import tempfile
from gettext import gettext as _

import editor
from easy_as_pypi_termio.errors import echo_warning

__all__ = (
    "ask_edit_with_editor",
    "run_editor_safe",
)


# ***


def ask_edit_with_editor(controller, fact=None, content=""):
    def _ask_edit_with_editor():
        contents = prepare_contents(content)
        filename = temp_filename()
        return run_editor_safe(filename, contents)

    def prepare_contents(content):
        content = content if content else ""
        # # FIXME: py2 compatible? Or need to six.b()?
        # #contents = six.b(str(content))  # NOPE: Has problems with Unicode, like: ½
        # contents = text_type(content).encode()
        # FIXME/2020-01-26: (lb): Verify no longer an issue.
        contents = str(content).encode()
        return contents

    def temp_filename():
        tmpfile = tempfile.NamedTemporaryFile(
            prefix=prepare_prefix(),
            suffix=prepare_suffix(),
        )
        filename = tmpfile.name
        return filename

    def prepare_prefix():
        # Vim names the terminal with the file's basename, which is
        # normally meaningless, e.g., "tmprvapy77w.rst (/tmp)", but
        # we can give the randomly-named temp file a prefix to make
        # the title more meaningful.
        prefix = None
        if fact is not None:
            # (lb): Reminder that colon is not acceptable for Windows paths.
            #   (I originally had a ':' in the clock time here.)
            # E.g., "2018_04_07_1733_"
            timefmt = "%Y_%m_%d_%H%M_"
            if fact.start:
                prefix = fact.start.strftime(timefmt)
            elif fact.end:
                prefix = fact.end.strftime(timefmt)
        return prefix

    def prepare_suffix():
        # User can set a suffix, which can be useful so, e.g., Vim
        # sees the extension and set filetype appropriately.
        # (lb): I like my Hamster logs to look like reST documents!
        suffix = controller.config["term.editor_suffix"] or None
        return suffix

    return _ask_edit_with_editor()


# ***


def run_editor_safe(filename, contents=None):
    def _run_editor_safe():
        try:
            return run_editor()
        except Exception as err:
            msg = _("Unable to run $EDITOR: {}").format(str(err))
            echo_warning(msg)
            return ""

    def run_editor():
        if is_editor_set() or not running_windows():
            # If Linux and EDITOR not set, editor.edit runs Vim.
            return run_editor_normal()
        else:
            return run_editor_windows()

    def is_editor_set():
        try:
            return bool(os.environ["EDITOR"])
        except KeyError:
            return False

    def running_windows():
        return platform.system() == "Windows"

    def run_editor_normal():
        # NOTE: You'll find EDITOR features in multiple libraries.
        #       The UX should be indistinguishable to the user.
        #       E.g., we could use click's `edit` instead of editor's:
        #
        #           click.edit(text=None,
        #                      editor=None,
        #                      env=None,
        #                      require_save=True,
        #                      extension='.txt',
        #                      filename=None)
        #
        #       Except that dob-viewer does not have click as a dependency.
        #
        # NOTE: Neither editor.edit nor click.edit appreciate arguments, e.g.,
        #       you might want to start Vim in insert mode and send cursor home:
        #
        #           export EDITOR="vim -c 'startinsert' -c 'norm! gg'"
        #
        #       but this'll crash (they treat the complete $EDITOR string as a
        #       path). (lb): But I'm not wrong expecting this! It works fine in
        #       other tools, e.g., `git commit -v` is perfectly happy with it.
        #
        #       As a work-around, you can put the command in an executable on
        #       PATH, e.g.,
        #
        #           echo -e '#!/bin/sh\nvim -c "startinsert" -c "norm! gg" "${@}"' \
        #               > ~/.local/bin/vim-wrap
        #           chmod 755 ~/.local/bin/vim-wrap
        #           export EDITOR="vim-wrap"
        #
        result = editor.edit(filename=filename, contents=contents)
        edited = result.decode()
        # Necessary?:
        #   edited = result.decode('utf-8')
        return edited

    def run_editor_windows():
        # NOTE: On Windows, EDITOR is not set by default, but neither is Vim, so
        #       editor.edit() will not work.
        #       - To set via PowerShell, try, e.g.,
        #           $Env:EDITOR = "notepad.exe"
        #       - In CMD.exe, use setx to set *persistent* variable (and then start
        #         another CMD prompt), e.g.,
        #           setx EDITOR "notepad.exe"
        # NOTE: There's a Windows-only os.startfile() that acts like double-clicking
        #       the file -- it opens the text file with the user's preferred editor --
        #       but it runs asynchronously. And we need to block. E.g., not this:
        #           os.startfile(filename, 'open')
        #       so just default to notepad, which will be installed. User should set
        #       EDITOR if they want to use a different text editor on Windows.
        with open(filename, "wb") as temp_f:
            temp_f.write(contents)
        import subprocess

        subprocess.call(["notepad.exe", filename])
        with open(filename, "r") as temp_f:
            edited = temp_f.read()
        return edited

    return _run_editor_safe()



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Controller functionality: Terminal *color* enablement."""

from easy_as_pypi_termio.style import disable_colors, enable_colors

from .pdb_subprocess import Controller_PdbSubprocess

__all__ = ("Controller_SetupColoring",)


class Controller_SetupColoring(
    # For self.config, depends on:
    #  Controller_ConfigManager  # self.config['term.*']
    # but for deliberate super() chaining: load data store layer next:
    Controller_PdbSubprocess,
):
    """Enables/Disables terminal output coloring per config's term.use_color."""

    def __init__(self, *args, **kwargs):
        super(Controller_SetupColoring, self).__init__(*args, **kwargs)

    # ***

    def setup_tty_color(self, use_color):
        if use_color is None:
            use_color = self.config["term.use_color"]
        else:
            self.config["term.use_color"] = use_color
        if use_color:
            enable_colors()
        else:
            disable_colors()

    # ***



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from config_decorator.config_decorator import ConfigDecorator
from easy_as_pypi_config.dec_wrap import decorate_and_wrap
from easy_as_pypi_config.fileboss import create_configobj, echo_config_obj
from easy_as_pypi_termio.echoes import click_echo, highlight_value
from easy_as_pypi_termio.errors import echo_warning, exit_warning
from easy_as_pypi_termio.style import attr

from ..config.config_table import echo_config_decorator_table
from ..crud.interrogate import run_editor_safe
from .create_conf import create_basic_conf
from .load_styling import load_rules_conf, load_style_rules, resolve_path_rules
from .rules_conf import create_style_rules_object
from .style_engine import StyleEngine

__all__ = (
    "create_rules_conf",
    "echo_rules_conf",
    "echo_rule_names",
    "echo_rules_table",
    "edit_rules_conf",
)


# *** [CONF] RULES


def echo_rules_conf(controller, rule_name, complete=False):
    config = controller.config

    def _echo_rules_conf():
        rules_confobj = load_rules_confobj()
        if rules_confobj:
            echo_config_obj(rules_confobj)
        # Else, already printed error message.

    def load_rules_confobj():
        rules_confobj, failed = load_rules_conf(config)
        if rules_confobj:
            return filter_rules_confobj(rules_confobj)
        if failed:
            # load_styles_conf prints a ConfigObj error message. Our job is done.
            return None
        return echo_error_no_rules_conf()

    def filter_rules_confobj(rules_confobj):
        if not rule_name:
            return rules_confobj
        new_config = create_configobj(conf_path=None)
        try:
            new_config.merge({rule_name: rules_confobj[rule_name]})
        except KeyError:
            return echo_error_no_rules_section(rule_name)
        else:
            return new_config

    def echo_error_no_rules_conf():
        msg = _("No rules file at: {0}").format(resolve_path_rules(config))
        echo_warning(msg)
        return None

    def echo_error_no_rules_section(rule_name):
        msg = _("No matching section “{0}” found in rules file at: {1}").format(
            rule_name,
            resolve_path_rules(config),
        )
        echo_warning(msg)
        return None

    # ***

    return _echo_rules_conf()


# *** [CREATE] RULES


def create_rules_conf(controller, force):
    def _create_rules_conf():
        object_name = _("Rules file")
        rules_path = resolve_path_rules(controller.config)
        create_basic_conf(rules_path, object_name, create_rules_file, force)

    def create_rules_file(rules_path):
        # Load specified style, or DEFAULT_STYLE if not specified.
        ruleset = create_style_rules_object()
        rule_name = _("Example Style Rule - Showing all built-in options")
        config_obj = decorate_and_wrap(rule_name, ruleset, complete=True)
        config_obj.filename = rules_path
        config_obj.write()

    _create_rules_conf()


# *** [EDIT] RULES


def edit_rules_conf(controller):
    rules_path = resolve_path_rules(controller.config)
    run_editor_safe(filename=rules_path)


# *** [LIST] RULES


def echo_rule_names(controller):
    """"""

    def _echo_rule_names():
        rules_confobj = load_style_rules(controller)
        print_rules_names(rules_confobj, _("User-created rules"))

    def print_rules_names(rules_confobj, title):
        click_echo("{}{}{}".format(attr("underlined"), title, attr("reset")))
        for rule_name in rules_confobj.keys():
            click_echo("  " + highlight_value(rule_name))

    return _echo_rule_names()


# *** [SHOW] RULES


def echo_rules_table(controller, name, output_format):
    def _echo_rules_table():
        if not name:
            rule_name, ruleset = create_example_rule()
        else:
            rule_name, ruleset = fetch_existing_rule()
        print_ruleset_table(rule_name, ruleset)

    def create_example_rule():
        rule_name = _("example")
        ruleset = create_style_rules_object()
        return rule_name, ruleset

    def fetch_existing_rule():
        rules_confobj = load_style_rules(controller)
        styling_rules = StyleEngine(rules_confobj)
        try:
            ruleset = styling_rules.rulesets[name]
        except KeyError:
            exit_rule_unknown(name)
        return name, ruleset

    def exit_rule_unknown(rule_name):
        exit_warning(_("No rule named “{}”").format(rule_name))

    def print_ruleset_table(rule_name, ruleset):
        condec = ConfigDecorator.create_root_for_section(rule_name, ruleset)
        conf_objs = [condec]
        echo_config_decorator_table(
            cfg_decors=conf_objs,
            exclude_section=False,
            # Passed on to render_results:
            controller=controller,
            output_format=output_format,
        )

    _echo_rules_table()



# This file exists within 'dob-bright':
#
#   https://github.com/tallybark/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Controller functionality: Database *store* features."""

import os
import sys
from gettext import gettext as _

from easy_as_pypi_termio.echoes import click_echo, highlight_value
from easy_as_pypi_termio.errors import exit_warning
from nark.items.fact import Fact

from .setup_coloring import Controller_SetupColoring

__all__ = ("Controller_DatabaseStore",)


class Controller_DatabaseStore(
    # For self.config, depends on:
    #  Controller_ConfigManager  # self.config['db.*']
    # For self.arg0name, depends on:
    #  Controller_PdbSubprocess
    # but for deliberate super() chaining: load data store layer next:
    Controller_SetupColoring,
):
    """Helps manage NarkControl.store for the CLI."""

    def __init__(self, *args, **kwargs):
        super(Controller_DatabaseStore, self).__init__(*args, **kwargs)

    @property
    def now(self):
        return self.store.now

    def now_refresh(self):
        return self.store.now_refresh()

    @property
    def data_store_exists_at(self):
        return _("Data store already exists at {}").format(self.config["db.path"])

    @property
    def data_store_url(self):
        return self.store.db_url

    @property
    def sqlite_db_path(self):
        if self.config["db.engine"] == "sqlite":
            return self.config["db.path"]

        # (lb): I don't super-like this. It's a weird side effect.
        #   And it's knowledgeable about the CLI command API. Meh.
        exit_warning(
            _("Not a SQLite database. Try `{} store url`").format(self.arg0name)
        )

    @property
    def is_germinated(self):
        if not self.store_exists:
            return False

        return Controller_SetupColoring.is_germinated.fget(self)

    @property
    def has_no_user_files(self):
        has_no_user_files = Controller_SetupColoring.has_no_user_files.fget(self)
        if has_no_user_files and not self.store_exists:
            return True

        return False

    @property
    def store_exists(self):
        # Check either db.path is set, or all of db.host/port/name/user.
        if self.config["db.engine"] == "sqlite":
            if self.config["db.path"] == ":memory:":
                return True
            return os.path.isfile(self.config["db.path"])
        else:
            # NOTE: db_url is an attribute of SQLAlchemyStore, not BaseStore.
            return bool(self.store.db_url)

    def standup_store(self, fact_cls=Fact):
        self.store.fact_cls = fact_cls
        return super(Controller_DatabaseStore, self).standup_store()

    # ***

    def create_data_store(self, force, fact_cls=Fact):
        unlinked_db = False
        skip_standup = self.check_sqlite_store_ready()
        if skip_standup:
            if force:
                self._reset_data_store()
                unlinked_db = True
            else:
                exit_warning(self.data_store_exists_at)
        self._standup_and_version_store(fact_cls)
        if unlinked_db:
            self._announce_recreated_store()

    # ***

    def check_sqlite_store_ready(self):
        if self.config["db.engine"] != "sqlite":
            return None
        db_path = self.config["db.path"]
        if not os.path.isfile(db_path):
            return False
        return True

    def _reset_data_store(self):
        if self.config["db.engine"] != "sqlite":
            # raise NotImplementedError
            exit_warning(_("FIXME: Reset non-SQLite data store not supported (yet)."))
        else:
            self.must_unlink_db_path(force=True)

    # ***

    def must_unlink_db_path(self, *_args, force):
        db_path = self.config["db.path"]
        if not os.path.exists(db_path):
            return
        if not os.path.isfile(db_path):
            exit_warning(
                _("Data store exists but is not a file, so not overwriting {}").format(
                    db_path
                )
            )
        if not force:
            exit_warning(self.data_store_exists_at)
        os.unlink(db_path)

    # ***

    def _announce_recreated_store(self):
        click_echo(
            _("Recreated data store at {}").format(
                highlight_value(self.config["db.path"])
            )
        )

    # ***

    def _standup_and_version_store(self, fact_cls):
        created_fresh = self.standup_store(fact_cls)
        if created_fresh:
            verb = _("created")
        else:
            verb = _("already ready")
        click_echo(
            _("Dob database {verb} at {url}").format(
                verb=verb,
                url=highlight_value(self.store.db_url),
            )
        )

    # ***

    def process_config(self):
        super(Controller_DatabaseStore, self).process_config()

        # *cough*hack!*cough*”
        # Because invoke_without_command, we allow command-less invocations.
        #   For one such call -- `dob -v` -- tell the store not to log.
        # Also tell the store not to log if user did not specify anything,
        #   because we'll show the help/usage (which Click would normally
        #   handle if we had not tampered with invoke_without_command).
        if (len(sys.argv) > 2) or (
            (len(sys.argv) == 2) and (sys.argv[1] not in ("-v", "version"))
        ):
            return
        # FIXME/EXPLAIN/2019-01-22: (lb): What about other 2 loggers?
        #   dev.cli_log_level
        #   dev.lib_log_level
        # (lb): Normally I'd prefer the []-lookup vs. attr., e.g., not:
        #   self.config.asobj.dev.sql_log_level.value_from_forced = 'WARNING'
        # because the self.config has non-key-val attributes (like
        # setdefault) so I think for clarity we should lookup via [].
        # Except the []-lookup returns the value, not the keyval object.
        # So here we have to use dotted attribute notation.
        self.config.asobj.dev.sql_log_level.value_from_forced = "WARNING"

    # ***
